# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re
from typing import Iterator

from click_extra.platforms import UNIX_WITHOUT_MACOS

from meta_package_manager.base import Package, PackageManager
from meta_package_manager.capabilities import version_not_implemented


class Emerge(PackageManager):
    """The Gentoo package manager.

    Documentation:
    - https://wiki.gentoo.org/wiki/Portage#emerge
    - https://dev.gentoo.org/~zmedico/portage/doc/man/emerge.1.html

    See other command equivalences at: https://wiki.archlinux.org/title/Pacman/Rosetta
    """

    homepage_url = "https://wiki.gentoo.org/wiki/Portage#emerge"

    platforms = UNIX_WITHOUT_MACOS

    requirement = "3.0.0"

    version_regex = r"Portage\s+(?P<version>\S+)"
    """
    .. code-block:: shell-session

        ► emerge --version
        Portage 3.0.30 (python 3.9.9-final-0, gcc-11.2.1, 5.15.32-gentoo-r1 x86_64)
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. warning::
            This suppose the ``qlist`` binary is available and present on the system. We
            do not search for it or try to resolves its canonical path with
            :py:attr:`PackageManager.cli_path
            <meta_package_manager.base.PackageManager.cli_path>`, as we do for the
            reference ``emerge`` binary.

        .. code-block:: shell-session

            ► qlist --installed --verbose --nocolor
            acct-group/audio-0-r1
            acct-group/cron-0
            app-admin/hddtemp-0.3_beta15-r29
            app-admin/perl-cleaner-2.30
            app-admin/system-config-printer-1.5.16-r1
            app-arch/p7zip-16.02-r8
        """
        # Locate qlist.
        qlist_path = self.which("qlist")
        if not qlist_path:
            raise FileNotFoundError(qlist_path)

        output = self.run_cli(
            "--installed",
            "--verbose",
            "--nocolor",
            override_cli_path=qlist_path,
        )

        regexp = re.compile(
            r"""
            (
                ?P<package_id>\S+   # Non-whitespace string...
                (?!-r)              # ...if not directly followred by the "-r" string.
            )
            -                       # A dash.
            (
                ?P<version>[^\s-]+  # Any non-whitespace/non-dash string.
                (?:-r\d+)?          # Optional revision suffix led by a -, non-grouped.
            )
            """,
            re.VERBOSE,
        )

        for package in output.splitlines():
            match = regexp.match(package)
            if match:
                package_id, installed_version = match.groups()
                yield self.package(id=package_id, installed_version=installed_version)

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            ► emerge --update --deep --pretend --columns --color n --nospinner @world
            [blocks  B     ] app-text/dos2unix
            [ebuild   N    ] app-games/qstat   [25c]
            [ebuild    R   ] sys-apps/sed      [2.4.7-r6]
            [ebuild       U] net-fs/samba      [2.2.8_pre1]      [2.2.7a]
            [ebuild       U] sys-devel/distcc  [2.16]            [2.13-r1] USE=ip6* -gtk
            [ebuild r     U] dev-libs/icu      [50.1.1:0/50.1.1] [50.1-r2:0/50.1]
            [ebuild r  R   ] dev-libs/libxml2  [2.9.0-r1:2]       USE=icu
        """
        output = self.run_cli(
            "--update",
            "--deep",
            "--pretend",
            "--columns",
            "--color",
            "n",
            "--nospinner",
            "@world",
        )

        regexp = re.compile(
            r"""
            \[.+\]                               # Update state.
            \                                    # A space.
            (?P<package_id>\S+)                  # Non-whitespace string.
            \s+                                  # Any spacing.
            (?:\[                                # Non-matching group
                                                 #   starting with a '['.
                (?P<latest_version>[^\s\/:]+)    # Any non-spaced string
                                                 #   until a ':' or '/' is met.
                \S*                              # Left-over parts of the version,
                                                 #   after a ':' or '/'.
            \])?                                 # Optional group ending with a ']'.
            \s+                                  # Any spacing.
            (?:\[                                # Non-matching group
                                                 #   starting with a '['.
                (?P<installed_version>[^\s\/:]+) # Any non-spaced string
                                                 #   until a ':' or '/' is met.
                \S*                              # Left-over parts of the version,
                                                 #   after a ':' or '/'.
            \])?                                 # Optional group ending with a ']'.
            """,
            re.VERBOSE,
        )

        for package in output.splitlines():
            match = regexp.match(package)
            if match:
                package_id, latest_version, installed_version = match.groups()
                yield self.package(
                    id=package_id,
                    latest_version=latest_version,
                    installed_version=installed_version,
                )

    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. code-block:: shell-session

            ► emerge --search --color n --nospinner blah

            [ Results for search key : blah ]
            Searching...

            *  sys-process/htop
                Latest version available: 1.0.2-r1
                Latest version installed: [ Not Installed ]
                Size of files: 380 KiB
                Homepage:      http://htop.sourceforge.net
                Description:   interactive process viewer
                License:       BSD GPL-2

            *  x11-drivers/nvidia-drivers
                Latest version available: 455.45.01-r1
                Latest version installed: [ Not Installed ]
                Size of files: 180.214 KiB
                Homepage:      https://www.nvidia.com/Download/Find.aspx
                Description:   NVIDIA Accelerated Graphics Driver
                License:       GPL-2 NVIDIA-r2

            [ Applications found : 2 ]

        .. code-block:: shell-session

            ► emerge --search --color n --nospinner %^sed$

        .. code-block:: shell-session

            ► emerge --searchdesc --color n --nospinner sed

        .. code-block:: shell-session

            ► emerge --searchdesc --color n --nospinner %^sed$
        """
        search_param = "--search"
        if extended:
            search_param = "--searchdesc"

        if exact:
            query = f"%^{query}$"

        output = self.run_cli(search_param, "--color", "n", "--nospinner", query)

        regexp = re.compile(
            r"""
            ^\*\s+(?P<package_id>\S+)\n
            \s+Latest\ version\ available:\s+(?P<latest_version>\S+)\n
            (?:\s+.+\n)+?
            \s+Description:\s+(?P<description>.+)\n
            """,
            re.MULTILINE | re.VERBOSE,
        )

        for package_id, version, description in regexp.findall(output):
            yield self.package(
                id=package_id,
                description=description,
                latest_version=version,
            )

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            ► sudo emerge --color n --nospinner dev-vcs/git
        """
        return self.run_cli("--color", "n", "--nospinner", package_id, sudo=True)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            ► sudo emerge --update --newuse --deep --color n --nospinner @world
        """
        return self.build_cli(
            "--update",
            "--newuse",
            "--deep",
            "--color",
            "n",
            "--nospinner",
            "@world",
            sudo=True,
        )

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            ► sudo emerge --update --color n --nospinner dev-vcs/git
        """
        return self.build_cli(
            "--update",
            "--color",
            "n",
            "--nospinner",
            package_id,
            sudo=True,
        )

    def sync(self) -> None:
        """Sync package metadata.

        .. code-block:: shell-session

            ► sudo emerge --sync --color n --nospinner
        """
        self.run_cli("--sync", "--color", "n", "--nospinner", sudo=True)

    def cleanup(self) -> None:
        """Removes things we don't need anymore.

        An update is forced before calling the clean commands, as `pointed to by the
        emerge documentation
        <https://wiki.gentoo.org/wiki/Gentoo_Cheat_Sheet#Recommended_method>`_:

        > As a safety measure, depclean will not remove any packages unless *all*
        > required dependencies have been resolved. As a consequence, it is often
        > necessary to run `emerge --update --newuse --deep @world` prior to depclean.

        .. warning::
            This suppose the ``eclean`` binary is available and present on the system.
            We do not search for it or try to resolves its canonical path with
            :py:attr:`PackageManager.cli_path
            <meta_package_manager.base.PackageManager.cli_path>`, as we do for the
            reference ``emerge`` binary.

        .. code-block:: shell-session

            ► sudo emerge --update --newuse --deep --color n --nospinner @world
            ► sudo emerge --depclean
            ► sudo eclean distfiles
        """
        # Forces an upgrade first, as recommended by emerge documentation.
        self.upgrade()

        self.run_cli("--depclean", sudo=True)

        eclean_path = self.which("eclean")
        if eclean_path:
            self.run_cli("distfiles", override_cli_path=eclean_path, sudo=True)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


from __future__ import annotations

import pytest

from meta_package_manager.base import Operations

from .conftest import default_manager_ids
from .test_cli import CLISubCommandTests


@pytest.fixture()
def subcmd():
    # The details of backup operations are only logged at INFO level.
    return "--verbosity", "INFO", "backup"


class TestBackup(CLISubCommandTests):
    @staticmethod
    def evaluate_signals(mid, stdout, stderr):
        yield from (
            f"Dumping packages from {mid}..." in stderr,
            f"warning: {mid} does not implement {Operations.installed}" in stderr,
            # Common "not found" warning message.
            f"warning: Skip unavailable {mid} manager." in stderr,
        )

    def test_default_all_managers_output_to_console(self, invoke, subcmd):
        result = invoke(subcmd)
        assert result.exit_code == 0
        assert "Print installed package list to <stdout>" in result.stderr
        self.check_manager_selection(result)

    def test_output_to_console(self, invoke, subcmd):
        result = invoke(subcmd, "-")
        assert result.exit_code == 0
        assert "Print installed package list to <stdout>" in result.stderr
        self.check_manager_selection(result)

    def test_output_to_file(self, invoke, subcmd):
        result = invoke(subcmd, "mpm-packages.toml")
        assert result.exit_code == 0
        assert "mpm-packages.toml" in result.stderr
        self.check_manager_selection(result)

    @default_manager_ids
    def test_single_manager_file_output(self, manager_id, invoke, subcmd):
        result = invoke(
            "--verbosity", "INFO", f"--{manager_id}", subcmd, "mpm-packages.toml"
        )
        assert result.exit_code == 0
        assert "mpm-packages.toml" in result.stderr
        self.check_manager_selection(result, {manager_id})



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import copy
from random import shuffle

import pytest

from meta_package_manager.version import Token, TokenizedString, parse_version


def reverse_fixtures(table):
    """Utility method to reverse a tuple of tuples."""
    return tuple(map(tuple, map(reversed, table)))


@pytest.mark.parametrize("value", [0, 123, "0", "123", "abc", "ABC", "123abc"])
def test_token_allowed_instanciation(value):
    Token(value)


@pytest.mark.parametrize("value", [None, -1, "a-b-c", 1.0])
def test_token_unauthorized_instanciation(value):
    with pytest.raises(TypeError):
        Token(value)


eq_values = [
    (Token("2345"), Token("2345")),
    (Token("2345"), Token(2345)),
    (Token("02345"), Token(2345)),
    (Token("02345"), "02345"),
    (Token("2345"), 2345),
    (Token("2345"), "2345"),
    (Token("0"), "0"),
    (Token("0"), 0),
    (Token("000"), 0),
    (Token("abc"), "abc"),
    (Token("abc"), Token("abc")),
]


ne_values = [
    (Token("2345"), Token("45")),
    (Token("2345"), Token(45)),
    (Token("2345"), Token(0)),
    (Token("2345"), Token("0")),
    (Token("2"), -2),
    (Token("2"), 0),
    (Token("abc"), "def"),
    (Token("Z"), "z"),
    (Token("acb"), 123),
]


gt_values = [
    (Token("9999"), Token("12")),
    (Token("9999"), Token("000000099")),
    (Token(9999), Token("12")),
    (Token(9999), Token(12)),
    (Token(9999), "0"),
    (Token(9999), 0),
    (Token(0), -2),
    (Token("0"), -2),
    (Token("abc"), -2),
    (Token("abc"), "ab"),
    (Token("a"), "Z"),
    (Token("z"), "Z"),
    (Token("a"), Token("Z")),
]


lt_values = [
    (Token("12"), Token("9999")),
    (Token("0000012"), Token("9999")),
    (Token("12"), Token(9999)),
    (Token(12), Token(9999)),
    (Token("12"), 9999),
    (Token("0"), 9999),
    (Token("ab"), "abc"),
    (Token("Z"), "z"),
]


@pytest.mark.parametrize(("token", "value"), eq_values)
def test_token_eq(token, value):
    assert token == value


@pytest.mark.parametrize(("token", "value"), ne_values)
def test_token_ne(token, value):
    assert token != value


@pytest.mark.parametrize(("token", "value"), gt_values)
def test_token_gt(token, value):
    assert token > value


@pytest.mark.parametrize(("token", "value"), lt_values)
def test_token_lt(token, value):
    assert token < value


@pytest.mark.parametrize(("token", "value"), gt_values + eq_values)
def test_token_ge(token, value):
    assert token >= value


@pytest.mark.parametrize(("token", "value"), lt_values + eq_values)
def test_token_le(token, value):
    assert token <= value


def test_token_hash():
    assert hash(Token("9999")) == hash(Token("9999"))
    assert hash(Token("9999")) == hash(Token(9999))
    assert hash(Token(9999)) == hash(Token(9999))
    assert hash(Token("09999")) != hash(Token("9999"))
    assert hash(Token("09999")) != hash(Token(9999))


@pytest.mark.parametrize(
    "value",
    (None, 0, 123, -1, "0", "1.2.3", "abc", "A-B-C", "123   a bc \n"),
)
def test_tokenized_string_allowed_instanciation(value):
    TokenizedString(value)


@pytest.mark.parametrize(
    "value",
    (1.0, [1, 2, 3], (1, 2, 3), {1, 2, 3}, {"a": 1, "b": 2}),
)
def test_tokenized_string_unauthorized_instanciation(value):
    with pytest.raises(TypeError):
        TokenizedString(value)


def test_tokenized_string_hash():
    assert hash(TokenizedString("1.2.3")) == hash(TokenizedString("1.2.3"))
    assert hash(TokenizedString(9999)) == hash(TokenizedString(9999))
    assert hash(TokenizedString("9999")) == hash(TokenizedString(9999))
    assert hash(TokenizedString("09999")) != hash(TokenizedString("9999"))
    assert hash(TokenizedString("09999")) != hash(TokenizedString(9999))
    assert hash(TokenizedString("1.2.3", separator=".")) != hash(
        TokenizedString("1.2.3", separator="_"),
    )


def test_tokenized_string_noop_instanciation():
    assert TokenizedString(None) is None


def test_tokenized_string_idempotent_instanciation():
    tok1 = TokenizedString("1.2.3")
    tok2 = TokenizedString(tok1)
    assert tok1 is tok2
    assert hash(tok1) == hash(tok2)
    assert tok1 == tok2


def test_tokenized_string_deepcopy():
    tok1 = TokenizedString("1.2.3")
    tok2 = copy.deepcopy(tok1)
    assert tok1 is not tok2
    assert hash(tok1) == hash(tok2)
    assert tok1 == tok2


version_list = (
    ("r2917_1", ("r", 2917, 1)),
    (" r     29   \n  17   1 x  ", ("r", 29, 17, 1, "x")),
    ("2020.03.24", (2020, 3, 24)),
    ("4.2.1-5666.3", (4, 2, 1, 5666, 3)),
    ("4.2.1-5666..3", (4, 2, 1, 5666, 3)),
    ("4.2.1-5666.3.", (4, 2, 1, 5666, 3)),
    (".4.2.1-5666.3", (4, 2, 1, 5666, 3)),
    ("4.2.1--5666.3", (4, 2, 1, 5666, 3)),
    ("4.2.1-#-5666.3", (4, 2, 1, 5666, 3)),
    ("4.2.1-ABC-5666.3", (4, 2, 1, "abc", 5666, 3)),
    (
        "4.2.1-éèàçÇÉÈ²³¼ÀÁÂÃÄ—ÅËÍÑÒÖÜÝåïš™-5666.3",
        (4, 2, 1, "eeaccee", 231, 4, "aaaaae", "ainooeueyastm", 5666, 3),
    ),
    ("1.3_1", (1, 3, 1)),
    ("2.40.20161221.0239", (2, 40, 20161221, 239)),
    ("latest", ("latest",)),
    ("1.2beta5", (1, 2, "beta", 5)),
    ("2.8.18-x86_64", (2, 8, 18, "x", 86, 64)),
    ("1.8.0_112-b16", (1, 8, 0, 112, "b", 16)),
    ("3.6.9_build_4685", (3, 6, 9, "build", 4685)),
    ("1.8.6-124-g6cd4c31", (1, 8, 6, 124, "g", 6, "cd", 4, "c", 31)),
    ("0.9999999", (0, 9999999)),
    ("0.0.0", (0, 0, 0)),
    ("1.0+git71+c79e264-r0", (1, 0, "git", 71, "c", 79, "e", 264, "r", 0)),
    ("20190718-r0", (20190718, "r", 0)),
    ("0.0.0-development", (0, 0, 0, "development")),
    ("v0.9.4.3", ("v", 0, 9, 4, 3)),
    ("6.0.1-0ubuntu1", (6, 0, 1, 0, "ubuntu", 1)),
    ("4.6.0+git+20160126-2", (4, 6, 0, "git", 20160126, 2)),
    ("1:5.25-2ubuntu1", (1, 5, 25, 2, "ubuntu", 1)),
    ("1.18.4ubuntu1.1", (1, 18, 4, "ubuntu", 1, 1)),
    ("20160104ubuntu1", (20160104, "ubuntu", 1)),
    ("3.113+nmu3ubuntu4", (3, 113, "nmu", 3, "ubuntu", 4)),
    (
        "13.0.2,8:d4173c853231432f001e99d882",
        (13, 0, 2, 8, "d", 4173, "c", 853231432, "f", 1, "e", 99, "d", 882),
    ),
    ("1.01+20150706hgc3698e0+dfsg-2", (1, 1, 20150706, "hgc", 3698, "e", 0, "dfsg", 2)),
)


@pytest.mark.parametrize(("v_string", "v_tuple"), version_list)
def test_version_tokenizer(v_string, v_tuple):
    assert TokenizedString(v_string) == v_tuple


compared_gt = (
    ("1", None),
    ("2.0", "1.0"),
    ("0.1", "0"),
    ("0.1", "0.0"),
    ("0.1", "0.0.0.0.0"),
    # ("0.1", "0.beta2"),
    ("2.0", "1.0"),
    ("2.0", "1"),
    ("3.1.10", "3.1.9"),
    # ("6.2", "6.2.0"),
    # ("9.52", "9d"),
    ("8.2p1_1", "8.2p1"),
    ("3.12.0a4", "3.7.0"),
    # ("3.12.0", "3.12.0a4"),
    # ("3.7.7", "3.7-beta2_1"),
    # ("2.1.1", "2.1.1-git-23hfb2-foobar"),
    (20200313, 20190801),
    ("2020.03.24", "2019.11.28"),
    # ("14.0.1,7:664493ef4a6946b186ff29eb326336a2",
    #  "14,36:076bab302c7b4508975440c56f6cc26a"),
)


@pytest.mark.parametrize(("ver1", "ver2"), compared_gt)
def test_version_comparison_gt(ver1, ver2):
    assert TokenizedString(ver1) > TokenizedString(ver2)
    assert parse_version(ver1) > parse_version(ver2)


@pytest.mark.parametrize(("ver1", "ver2"), reverse_fixtures(compared_gt))
def test_version_comparison_lt(ver1, ver2):
    assert TokenizedString(ver1) < TokenizedString(ver2)
    assert parse_version(ver1) < parse_version(ver2)


@pytest.mark.parametrize(
    "sequence",
    (
        [
            None,
            "r0",
            "r1",
            "r9_0",
            "r9_3",
            "r9_30",
            "r2917_1",
            "r02918_1",
            "r02920_0",
            "r02920_1",
            "r02920_20",
            "r02920_021",
        ],
    ),
)
def test_version_sorting(sequence):
    sorted_version = list(map(TokenizedString, sequence))
    ramdom_order = sorted_version.copy()
    shuffle(ramdom_order)
    assert ramdom_order != sorted_version
    assert sorted(ramdom_order) == sorted_version



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import logging
import os
import re
import shutil
import sys
from contextlib import nullcontext
from dataclasses import asdict, dataclass
from enum import Enum
from functools import cached_property
from pathlib import Path
from textwrap import dedent, indent, shorten
from typing import ContextManager, Generator, Iterable, Iterator, cast
from unittest.mock import patch
import subprocess

from boltons.iterutils import flatten, unique
from boltons.strutils import strip_ansi
from click_extra.colorize import default_theme as theme
from click_extra.platforms import UNIX, Group, Platform, current_os
from click_extra.testing import (
    INDENT,
    Arg,
    EnvVars,
    NestedArgs,
    args_cleanup,
    format_cli_prompt,
    env_copy,
)

from .version import TokenizedString, parse_version

Operations = Enum(
    "Operations",
    (
        "installed",
        "outdated",
        "search",
        "install",
        "upgrade",
        "upgrade_all",
        "remove",
        "sync",
        "cleanup",
    ),
)
"""Recognized operation IDs that are implemented by package manager with their specific
CLI invocation.

Each operation has its own CLI subcommand.
"""


class CLIError(Exception):
    """An error occurred when running package manager CLI."""

    def __init__(self, code: int, output: str, error: str) -> None:
        """The exception internally keeps the result of CLI execution."""
        super().__init__()
        self.code = code
        self.output = output
        self.error = error

    def __str__(self) -> str:
        """Human-readable error."""
        indented_output = indent(str(self.output), INDENT)
        indented_error = indent(str(self.error), INDENT)
        return indent(
            dedent(
                f"""
                Return code: {self.code}
                Output:
                {indented_output}
                Error:
                {indented_error}""",
            ),
            INDENT,
        )

    def __repr__(self) -> str:
        error_excerpt = shorten(
            " ".join(self.error.split()),
            width=60,
            placeholder="(...)",
        )
        return f"<{self.__class__.__name__}({self.code}, {error_excerpt!r})>"


@dataclass
class Package:
    """Lightweight representation of a package and its metadata."""

    id: str
    """ID is required and is the primary key used by the manager."""

    name: str | None = None

    description: str | None = None

    installed_version: TokenizedString | str | None = None
    latest_version: TokenizedString | str | None = None
    """Installed and latest versions are optional: they're not always provided by the
    package manager.

    ``installed_version`` and ``latest_version`` are allowed to temporarily be strings
    between ``__init__`` and ``__post_init__``. Once they reach the later, they're
    parsed and normalized into either ``TokenizedString`` or `None`. They can't be
    strings beyong that point, i.e. after the Package instance has been fully
    instantiated. We don't know how to declare this transient state with type hints,
    so we're just going to allow string type.
    """

    def __post_init__(self) -> None:
        # Make sure version strings are parsed into proper objects.
        self.installed_version = parse_version(self.installed_version)
        self.latest_version = parse_version(self.latest_version)


def packages_asdict(packages: Iterator[Package], keep_fields: tuple[str, ...]):
    """Returns a list of packages casted to a ``dict`` with only a subset of its
    fields."""
    return ({k: v for k, v in asdict(p).items() if k in keep_fields} for p in packages)


class MetaPackageManager(type):
    """Custom metaclass used as a class factory for package managers."""

    def __init__(cls, name, bases, dct) -> None:
        """Sets some class defaults, but only if they're not redefined in the final
        manager class.

        Also normalize list of platform, by ungrouping groups, deduplicate entries and
        freeze them into a set of unique platforms.
        """
        if "id" not in dct:
            cls.id = name.lower().replace("_", "-")

        if "name" not in dct:
            cls.name = name

        if "cli_names" not in dct:
            cls.cli_names = (cls.id,)

        if "virtual" not in dct:
            cls.virtual = name == "PackageManager" or not cls.cli_names

        if "platforms" in dct:
            platforms = set()
            for spec in flatten([dct["platforms"]]):
                if isinstance(spec, Platform):
                    platforms.add(spec)
                elif isinstance(spec, Group):
                    platforms |= set(spec.platforms)
                else:
                    msg = f"Unrecognized platform specifier: {spec!r}"
                    raise ValueError(msg)
            cls.platforms = frozenset(platforms)


def highlight_cli_name(path: Path | None, match_names: Iterable[str]) -> str | None:
    """Highlight the binary name in the provided ``path``.

    If ``match_names`` is provided, only highlight the start of the binary name that is
    in the list.

    Matchin is insensitive to case on Windows and case-sensitive on other platforms,
    thanks to ``os.path.normcase``.
    """
    if path is None:
        return None

    highlighted_name = path.name
    for ref_name in match_names:
        if os.path.normcase(ref_name).startswith(os.path.normcase(path.name)):
            highlighted_name = (
                theme.invoked_command(path.name[: len(ref_name)])
                + path.name[len(ref_name) :]
            )
            break

    return f"{path.parent}{os.path.sep}{highlighted_name}"


class PackageManager(metaclass=MetaPackageManager):
    """Base class from which all package manager definitions inherits."""

    deprecated: bool = False
    """A manager marked as deprecated will be hidden from all package selection by
    default.

    You can still use it but need to explicitly call for it on the command line.

    Implementation of a deprecated manager will be kept within mpm source code, but some
    of its features or total implementation are allowed to be scraped in the face of
    maintenance pain and adversity.

    Integration tests and unittests for deprecated managers can be removed. We do not
    care if a deprecated manager is not 100% reliable. A flakky deprecated manager
    should not block a release due to flakky tests.
    """

    deprecation_url: str | None = None
    """Announcement from the official project or evidence of abandonment of
    maintainance."""

    id: str
    """Package manager's ID.

    Derived by defaults from the lower-cased class name in which underscores ``_`` are
    replaced by dashes ``-``.

    This ID must be unique among all package manager definitions and lower-case, as
    they're used as feature flags for the :program:`mpm` CLI.
    """

    name: str
    """Return package manager's common name.

    Default value is based on class name.
    """

    homepage_url: str | None = None
    """Home page of the project, only used in documentation for reference."""

    platforms: Group | Platform | Iterable[Group | Platform] | frozenset[Platform] = (
        frozenset()
    )
    """List of platforms supported by the manager.

    Allows for a mishmash of platforms and groups. Will be normalized into a frozen set
    of ``Platform`` instances at instanciation.
    """

    requirement: str | None = None
    """Minimal required version.

    Should be a string parseable by
    :py:class:`meta_package_manager.version.parse_version`.

    Defaults to ``None``, which deactivate version check entirely.
    """

    cli_names: tuple[str, ...]
    """List of CLI names the package manager is known as.

    The supported CLI names are ordered by priority. This is used for example to help
    out the search of the right binary in the case of the python3/python2 transition.

    Is derived by default from the manager's ID.
    """

    virtual: bool
    """Should we expose the package manager to the user?

    Virtual package manager are just skeleton classes used to factorize code among
    managers of the same family.
    """

    cli_search_path: tuple[str, ...] = ()
    """List of additional path to help :program:`mpm` hunt down the package manager CLI.

    Must be a list of strings whose order dictates the search sequence.

    Most of the time unnecessary:
    :py:func:`meta_package_manager.base.PackageManager.cli_path` works well on all
    platforms.
    """

    extra_env: EnvVars | None = None
    """Additional environment variables to add to the current context.

    Automatically applied on each
    :py:func:`meta_package_manager.base.PackageManager.run_cli` calls.
    """

    pre_cmds: tuple[str, ...] = ()
    """Global list of pre-commands to add before before invoked CLI.

    Automatically added to each
    :py:func:`meta_package_manager.base.PackageManager.run_cli` call.

    Used to prepend `sudo <https://www.sudo.ws>`_ or other system utilities.
    """

    pre_args: tuple[str, ...] = ()
    post_args: tuple[str, ...] = ()
    """Global list of options used before and after the invoked package manager CLI.

    Automatically added to each
    :py:func:`meta_package_manager.base.PackageManager.run_cli` call.

    Essentially used to force silencing, low verbosity or no-color output.
    """

    version_cli_options: tuple[str, ...] = ("--version",)
    """List of options to get the version from the package manager CLI."""

    version_regex: str = r"(?P<version>\S+)"
    """Regular expression used to extract the version number from the result of the CLI
    run with the options above. It doesn't matter if the regex returns unsanitized and
    crappy string. The :py:func:`meta_package_manager.base.PackageManager.version`
    method will clean and normalized it.

    By default match the first part that is space-separated.
    """

    stop_on_error: bool = False
    """Tell the manager to either raise or continue on errors."""

    ignore_auto_updates: bool = True
    """Some managers can report or ignore packages which have their own auto-update
    mechanism."""

    dry_run: bool = False
    """Do not actually perform any action, just simulate CLI calls."""

    timeout: int | None = None
    """Maximum number of seconds to wait for a CLI call to complete."""

    cli_errors: list[CLIError]
    """Accumulate all CLI errors encountered by the package manager."""

    package: type[Package] = Package
    """The dataclass to use to produce Package objects from the manager."""

    def __init__(self) -> None:
        """Initialize ``cli_errors`` list."""
        self.cli_errors = []

    @classmethod
    def implements(cls, op: Operations) -> bool:
        """Inspect manager's implementation to check for proper support of an
        operation."""
        logging.debug(f"Does {cls} implements {op}?")

        # General case: the operation and the method implementing it shares the same ID.
        method_deps: tuple[set[str], ...] = ({op.name},)

        # Special case for single-package `upgrade`: we depends on `upgrade_one_cli()`.
        if op == Operations.upgrade:
            method_deps = ({"installed", "upgrade_one_cli"},)

        # For `upgrade_all`: we depends on eother `upgrade_all_cli()`, or we can
        # simulate the latter with a combination of `outdated()` and
        # `upgrade_one_cli()`.
        elif op == Operations.upgrade_all:
            method_deps = ({"upgrade_all_cli"}, {"outdated", "upgrade_one_cli"})

        # If none of the classes in the inheritance hierarchy up to the base one
        # implements the operation, then we can be certain the manager doesn't implement
        # the operation at all.
        for klass in cls.mro():
            if klass is PackageManager:
                return False
            # Presence of the operation function is not enough to rules out proper
            # implementation, as it can be a method that raises NotImplemented error
            # anyway. See for instance the upgrade_all_cli in pip.py:
            # https://github.com/kdeldycke/meta-package-manager/blob/4acc003/meta_package_manager/managers/pip.py#L271-L279
            for method_ids in method_deps:
                all_deps_found = method_ids.issubset(klass.__dict__)
                if all_deps_found:
                    return True

        msg = f"Can't guess {cls} implementation of {op}."
        raise NotImplementedError(msg)

    def search_all_cli(
        self,
        cli_names: Iterable[str],
        env=None,
    ) -> Generator[Path, None, None]:
        """Search for all binary files matching the CLI names, in all environment path.

        This is like our own implementation of ``shutil.which()``, with the difference
        that it is capable of returning all the possible paths of the provided file
        names, in all environment path, not just the first one that match. And on
        Windows, prevents matching of CLI in the current directory, which takes
        precedence on other paths.

        Returns all files matching any ``cli_names``, by iterating over all folders in
        this order:
          * folders provided by :py:attr:`cli_search_path
            <meta_package_manager.base.PackageManager.cli_search_path>`,
          * then in all the default places specified by the environment variable (i.e.
            ``os.getenv("PATH")``).

        Only returns files that exists and are not empty.

        .. caution::

            Symlinks are not resolved, because some manager like `Homebrew on Linux
            relies on some sort of symlink-based trickery
            <https://github.com/kdeldycke/meta-package-manager/pull/188>`_ to set
            environment variables.
        """
        # Check CLI names are not path, but plain filenames.
        for cli_name in cli_names:
            assert not os.path.dirname(
                cli_name,
            ), f"CLI name {cli_name} contains path separator {os.path.sep}."

        # Validates each search path.
        for cli_search_path in self.cli_search_path:
            assert os.pathsep not in cli_search_path, (
                f"Search path {cli_search_path} contains "
                f"environment path separator {os.pathsep}."
            )

        # By default, the filename to search for is the case-sensitive CLI name.
        search_filenames = list(cli_names)
        # But on Windows, there is this special ``PATHEXT`` environment variable to
        # tell you what file suffixes are executable. We have to search for any
        # variation of the CLI name with any of these suffixes.
        # Code below is inpired by the original implementation of ``shutil.which()``:
        # https://github.com/python/cpython/blob/8d46c7e/Lib/shutil.py#L1478-L1491
        if sys.platform == "win32":
            pathext_source = os.getenv("PATHEXT") or shutil._WIN_DEFAULT_PATHEXT
            pathext = unique(ext for ext in pathext_source.split(os.pathsep) if ext)
            search_filenames = []
            for cli_name in cli_names:
                # See if the given file matches any of the expected path extensions.
                # This will allow us to short circuit when given "python.exe".
                # If it does match, only test that one, otherwise we have to try
                # others.
                if any(cli_name.lower().endswith(ext.lower()) for ext in pathext):
                    search_filenames.append(cli_name)
                else:
                    search_filenames.extend(f"{cli_name}{ext}" for ext in pathext)
        search_filenames = unique(search_filenames)

        def normalize_path(path: Path) -> str:
            """Resolves symlinks and produces a normalized absolute path string.

            Additonnaly use ``os.path.normcase`` on Windows to exclude duplicates
            produced by case-insensitive filesystems.
            """
            return os.path.normcase(path.resolve())

        # Deduplicate search paths while keeping their order and original value, as the
        # normalization process happens with the ``key`` lookup.
        search_path_list: list[Path] = unique(
            # Manager-specific search path takes precedence over default environment.
            (Path(p) for p in list(self.cli_search_path) + os.get_exec_path(env=env)),
            key=normalize_path,
        )

        logging.debug(
            "Search for "
            + ", ".join(theme.invoked_command(cli) for cli in search_filenames)
            + " in "
            + ", ".join(str(p) for p in search_path_list),
        )

        for search_path in search_path_list:
            if not search_path.is_dir():
                continue

            for filename in search_filenames:
                file = search_path / filename
                if not file.is_file() or not os.path.getsize(file):
                    continue
                logging.debug(f"CLI found at {highlight_cli_name(file, cli_names)}")
                yield file

    def which(self, cli_name: str) -> Path | None:
        """Emulates the ``which`` command.

        Based on the ``search_all_cli()`` method.
        """
        for cli_path_found in self.search_all_cli([cli_name]):
            return cli_path_found
        return None

    @cached_property
    def cli_path(self) -> Path | None:
        """Fully qualified path to the canonical package manager binary.

        Try each CLI names provided by :py:attr:`cli_names
        <meta_package_manager.base.PackageManager.cli_names>`, in each system path
        provided by :py:attr:`cli_search_path
        <meta_package_manager.base.PackageManager.cli_search_path>`. In that order.
        Then returns the first match.

        Executability of the CLI will be separately assessed later by the
        :py:func:`meta_package_manager.base.PackageManager.executable` method below.
        """
        if self.cli_names is not None:
            for cli_path in self.search_all_cli(self.cli_names):
                return cli_path
        return None

    @cached_property
    def version(self) -> TokenizedString | None:
        """Invoke the manager and extract its own reported version string.

        Returns a parsed and normalized version in the form of a
        :py:class:`meta_package_manager.version.TokenizedString` instance.
        """
        if self.executable:
            # Invoke the manager.
            try:
                output = self.run_cli(
                    self.version_cli_options,
                    auto_pre_cmds=False,
                    auto_pre_args=False,
                    auto_post_args=False,
                    force_exec=True,
                )
            # Catch false-positive CLIs.
            except OSError as ex:
                # In the environment on Windows, extension of available executables are
                # ignored as they're plenty: .EXE, .CMD, .BAT, ...
                # See: https://github.com/kdeldycke/meta-package-manager/issues/542
                # Check for "OSError: [WinError 193] %1 is not a valid Win32
                # application" error.
                if getattr(ex, "winerror", None) == 193:
                    logging.debug(
                        f"{theme.invoked_command(str(self.cli_path))} "
                        "is not a valid Windows application.",
                    )
                    # Declare CLI as un-executable.
                    self.executable = False
                    return None
                # Unidentified error: re-raise.
                raise

            # Extract the version with the regex.
            parts = re.compile(self.version_regex, re.MULTILINE | re.VERBOSE).search(
                output,
            )
            if parts:
                version_string = parts.groupdict().get("version")
                logging.debug(f"Extracted version: {version_string!r}")
                parsed_version = parse_version(version_string)
                logging.debug(f"Parsed version: {parsed_version!r}")
                return parsed_version
        return None

    @cached_property
    def supported(self) -> bool:
        """Is the package manager supported on that platform?"""
        # At this point self.platforms is normalized as a frozenset of Platform
        # instances, but mypy doens't know it.
        return current_os() in self.platforms  # type: ignore[operator]

    @cached_property
    def executable(self) -> bool:
        """Is the package manager CLI can be executed by the current user?"""
        if not self.cli_path:
            return False
        if not os.access(self.cli_path, os.X_OK):
            logging.debug(
                f"{highlight_cli_name(self.cli_path, self.cli_names)} "
                "is not allowed to be executed.",
            )
            return False
        return True

    @cached_property
    def fresh(self) -> bool:
        """Does the package manager match the version requirement?"""
        # Version is mandatory.
        if not self.version:
            return False
        if self.requirement and self.version < parse_version(self.requirement):
            logging.debug(
                f"{self.id} {self.version} is older than "
                f"{self.requirement} version requirement.",
            )
            return False
        return True

    @cached_property
    def available(self) -> bool:
        """Is the package manager available and ready-to-use on the system?

        Returns ``True`` only if the main CLI:

        1. is :py:attr:`supported on the current platform
           <meta_package_manager.base.PackageManager.supported>`,
        2. was :py:attr:`found on the system
           <meta_package_manager.base.PackageManager.cli_path>`,
        3. is :py:attr:`executable
           <meta_package_manager.base.PackageManager.executable>`, and
        4. :py:attr:`match the version requirement
           <meta_package_manager.base.PackageManager.fresh>`.
        """
        logging.debug(
            f"{self.id} "
            f"is deprecated: {self.deprecated}; "
            f"is supported: {self.supported}; "
            f"found at: {highlight_cli_name(self.cli_path, self.cli_names)}; "
            f"is executable: {self.executable}; "
            f"is fresh: {self.fresh}.",
        )
        return bool(self.supported and self.cli_path and self.executable and self.fresh)

    def run(self, *args: Arg | NestedArgs, extra_env: EnvVars | None = None) -> str:
        """Run a shell command, return the output and accumulate error messages.

        ``args`` is allowed to be a nested structure of iterables, in which case it will
        be recursively flatten, then ``None`` will be discarded, and finally each item
        casted to strings.

        Running commands with that method takes care of:
          * adding logs at the appropriate level
          * removing ANSI escape codes from
            :py:attr:`subprocess.CompletedProcess.stdout` and
            :py:attr:`subprocess.CompletedProcess.stderr`
          * returning ready-to-use normalized strings (dedented and stripped)
          * letting :option:`mpm --dry-run` and :option:`mpm --stop-on-error` have
            expected effect on execution
        """
        # Casting to string helps serialize Path and Version objects.
        clean_args = args_cleanup(*args)
        cli_msg = format_cli_prompt(clean_args, extra_env)

        code = 0
        output = ""
        error = ""

        if self.dry_run:
            logging.warning(f"Dry-run: {cli_msg}")
        else:
            logging.debug(cli_msg)
            result = subprocess.run(
                clean_args,
                capture_output=True,
                timeout=self.timeout,
                encoding="utf-8",
                env=cast("subprocess._ENV", env_copy(extra_env)),
            )
            code = result.returncode
            output = result.stdout
            error = result.stderr

        # Normalize messages.
        if error:
            error = dedent(strip_ansi(error).strip())
        if output:
            output = dedent(strip_ansi(output).strip())

        # Log <stdout> and <stderr> output.
        if output:
            logging.debug(indent(output, INDENT))
        if error:
            # Non-fatal error messages are logged as warnings.
            log_func = logging.error if code else logging.warning
            log_func(indent(error, INDENT))

        # Non-successful run.
        if code and error:
            # Produce an exception and eventually raise it.
            exception = CLIError(code, output, error)
            if self.stop_on_error:
                raise exception
            # Accumulate errors.
            self.cli_errors.append(exception)

        return output

    def build_cli(
        self,
        *args: Arg | NestedArgs,
        auto_pre_cmds: bool = True,
        auto_pre_args: bool = True,
        auto_post_args: bool = True,
        override_pre_cmds: NestedArgs | None = None,
        override_cli_path: Path | None = None,
        override_pre_args: NestedArgs | None = None,
        override_post_args: NestedArgs | None = None,
        sudo: bool = False,
    ) -> tuple[str, ...]:
        """Build the package manager CLI by combining the custom ``*args`` with the
        package manager's global parameters.

        Returns a tuple of strings.

        Helps the construction of CLI's repeating patterns and makes the code easier to
        read. Just pass the specific ``*args`` and the full CLI string will be composed
        out of the globals, following this schema:

        .. code-block:: shell-session

            $ [<pre_cmds>|sudo] <cli_path> <pre_args> <*args> <post_args>

        * :py:attr:`self.pre_cmds <meta_package_manager.base.PackageManager.pre_cmds>`
          is added before the CLI path.

        * :py:attr:`self.cli_path <meta_package_manager.base.PackageManager.cli_path>`
          is used as the main binary to execute.

        * :py:attr:`self.pre_args <meta_package_manager.base.PackageManager.pre_args>`
          and :py:attr:`self.post_args
          <meta_package_manager.base.PackageManager.post_args>`  globals are added
          before and after the provided ``*args``.

        Each additional set of elements can be disabled with their respective flag:

        * ``auto_pre_cmds=False``  to skip the automatic addition of
          :py:attr:`self.pre_cmds <meta_package_manager.base.PackageManager.pre_cmds>`
        * ``auto_pre_args=False``  to skip the automatic addition of
          :py:attr:`self.pre_args <meta_package_manager.base.PackageManager.pre_args>`
        * ``auto_post_args=False`` to skip the automatic addition of
          :py:attr:`self.post_args <meta_package_manager.base.PackageManager.post_args>`

        Each global set of elements can be locally overridden with:

        * ``override_pre_cmds=tuple()``
        * ``override_cli_path=str``
        * ``override_pre_args=tuple()``
        * ``override_post_args=tuple()``

        On linux, the command can be run with `sudo <https://www.sudo.ws>`_ if the
        parameter of the same name is set to ``True``. In which case the
        ``override_pre_cmds`` parameter is not allowed to be set and the
        ``auto_pre_cmds`` parameter is forced to ``False``.
        """
        params: list[Arg | NestedArgs] = []

        # Sudo replaces any pre-command, be it overridden or automatic.
        if sudo:
            if current_os() not in UNIX:
                msg = "sudo only supported on UNIX."
                raise NotImplementedError(msg)
            if override_pre_cmds:
                msg = "Pre-commands not allowed if sudo is requested."
                raise ValueError(msg)
            if auto_pre_cmds:
                auto_pre_cmds = False
            params.append("sudo")
        elif override_pre_cmds:
            params.extend(override_pre_cmds)  # type: ignore[arg-type]
        elif auto_pre_cmds:
            params.extend(self.pre_cmds)

        if override_cli_path:
            params.append(override_cli_path)
        else:
            params.append(self.cli_path)

        if override_pre_args:
            params.extend(override_pre_args)  # type: ignore[arg-type]
        elif auto_pre_args:
            params.extend(self.pre_args)

        if args:
            params.extend(args)

        if override_post_args:
            params.extend(override_post_args)  # type: ignore[arg-type]
        elif auto_post_args:
            params.extend(self.post_args)

        return args_cleanup(params)  # type: ignore[arg-type]

    def run_cli(
        self,
        *args: Arg | NestedArgs,
        auto_extra_env: bool = True,
        auto_pre_cmds: bool = True,
        auto_pre_args: bool = True,
        auto_post_args: bool = True,
        override_extra_env: EnvVars | None = None,
        override_pre_cmds: NestedArgs | None = None,
        override_cli_path: Path | None = None,
        override_pre_args: NestedArgs | None = None,
        override_post_args: NestedArgs | None = None,
        force_exec: bool = False,
        sudo: bool = False,
    ) -> str:
        """Build and run the package manager CLI by combining the custom ``*args`` with
        the package manager's global parameters.

        After the CLI is built with the
        :py:meth:`meta_package_manager.base.PackageManager.build_cli` method, it is
        executed with the :py:meth:`meta_package_manager.base.PackageManager.run`
        method, augmented with environment variables from :py:attr:`self.extra_env
        <meta_package_manager.base.PackageManager.extra_env>`.

        All parameters are the same as
        :py:meth:`meta_package_manager.base.PackageManager.build_cli`, plus:

        * ``auto_extra_env=False`` to skip the automatic addition of
          :py:attr:`self.extra_env <meta_package_manager.base.PackageManager.extra_env>`
        * ``override_extra_env=dict()`` to locally overrides the later
        * ``force_exec`` ignores the :option:`mpm --dry-run` and :option:`mpm
          --stop-on-error` options to force the execution and completion of the command.
        """
        cli = self.build_cli(
            *args,
            auto_pre_cmds=auto_pre_cmds,
            auto_pre_args=auto_pre_args,
            auto_post_args=auto_post_args,
            override_pre_cmds=override_pre_cmds,
            override_cli_path=override_cli_path,
            override_pre_args=override_pre_args,
            override_post_args=override_post_args,
            sudo=sudo,
        )

        # Prepare the full list of CLI arguments.
        extra_env = None
        if override_extra_env:
            extra_env = override_extra_env
        elif auto_extra_env:
            extra_env = self.extra_env

        # No-op context manager without any effects.
        local_option1: ContextManager = nullcontext()
        local_option2: ContextManager = nullcontext()
        # Temporarily replace --dry-run and --stop-on-error user options with our own.
        if force_exec:
            local_option1 = patch.object(self, "dry_run", False)
            local_option2 = patch.object(self, "stop_on_error", False)
        # Execute the command with eventual local options.
        with local_option1, local_option2:
            return self.run(*cli, extra_env=extra_env)

    @property
    def installed(self) -> Iterator[Package]:
        """List packages currently installed on the system.

        Optional. Will be simply skipped by :program:`mpm` if not implemented.
        """
        raise NotImplementedError

    @property
    def outdated(self) -> Iterator[Package]:
        """List installed packages with available upgrades.

        Optional. Will be simply skipped by :program:`mpm` if not implemented.
        """
        raise NotImplementedError

    @classmethod
    def query_parts(cls, query: str) -> set[str]:
        """Returns a set of all contiguous alphanumeric string segments.

        Contrary to :py:class:`meta_package_manager.version.TokenizedString`, do no
        splits on colated number/alphabetic junctions.
        """
        return {p for p in re.split(r"\W+", query) if p}

    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Search packages available for install.

        There is no need for this method to be perfect and sensitive to ``extended`` and
        ``exact`` parameters. If the package manager is not supporting these kind of
        options out of the box, just returns the closest subset of matching package you
        can come up with. Finer refiltering will happens in the
        :py:meth:`meta_package_manager.base.PackageManager.refiltered_search` method
        below.

        Optional. Will be simply skipped by :program:`mpm` if not implemented.
        """
        raise NotImplementedError

    def refiltered_search(
        self,
        query: str,
        extended: bool,
        exact: bool,
    ) -> Iterator[Package]:
        """Returns search results with extra manual refiltering to refine gross
        matchings.

        Some package managers returns unbounded results, and/or don't support fine
        search criterions. In which case we use this method to manually refilters
        :py:meth:`meta_package_manager.base.PackageManager.search` results to either
        exclude non-extended or non-exact matches.

        Returns a generator producing the same data as the
        :py:meth:`meta_package_manager.base.PackageManager.search` method above.

        .. tip::

            If you are implementing a package manager definition, do not waste time to
            filter CLI results. Let this method do this job.

            Instead, just implement the core
            :py:meth:`meta_package_manager.base.PackageManager.search` method above and
            try to produce results as precise as possible using the native filtering
            capabilities of the package manager CLI.
        """
        for match in self.search(query, extended, exact):
            # Look by default into package ID and name.
            search_content = {match.id, match.name}

            # Rejects fuzzy results: only keep packages strictly matching on ID or name.
            if exact and query not in search_content:
                continue

            # Add description to the list of content to look into.
            if extended:
                search_content.add(match.description)

            # Normalize searched content.
            serialized_content = "".join({s.lower() for s in search_content if s})

            # Exclude packages not matching any part of the query.
            confirmed_match = False
            for part in {p.lower() for p in self.query_parts(query)}:
                if part in serialized_content:
                    confirmed_match = True
                    break
            if not confirmed_match:
                continue

            # Report the package as matching.
            yield match

    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package and one only.

        Allows a specific ``version`` to be provided.
        """
        raise NotImplementedError

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Returns the complete CLI to upgrade all outdated packages on the system."""
        raise NotImplementedError

    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Returns the complete CLI to upgrade one package and one only.

        Allows a specific ``version`` to be provided.
        """
        raise NotImplementedError

    def upgrade(self, package_id: str | None = None, version: str | None = None) -> str:
        """Perform an upgrade of either all or one package.

        Executes the CLI provided by either
        :py:meth:`meta_package_manager.base.PackageManager.upgrade_all_cli` or
        :py:meth:`meta_package_manager.base.PackageManager.upgrade_one_cli`.

        If the manager doesn't provides a full upgrade one-liner (i.e. if
        :py:meth:`meta_package_manager.base.PackageManager.upgrade_all_cli` raises
        :py:exc:`NotImplementedError`), then the list of all outdated packages will be
        fetched (via :py:meth:`meta_package_manager.base.PackageManager.outdated`) and
        each package will be updated one by one by calling
        :py:meth:`meta_package_manager.base.PackageManager.upgrade_one_cli`.

        See for example the case of
        :py:meth:`meta_package_manager.managers.pip.Pip.upgrade_one_cli`.
        """
        if package_id:
            cli = self.upgrade_one_cli(package_id, version=version)

        else:
            try:
                cli = self.upgrade_all_cli()
            except NotImplementedError:
                logging.info(
                    "Fallback to calling upgrade operation on each outdated package.",
                )
                log = ""
                for package in self.outdated:
                    output = self.upgrade(package.id)
                    if output:
                        log += f"\n{output}"
                return log

        return self.run(cli, extra_env=self.extra_env)

    def remove(self, package_id: str) -> str:
        """Remove one package and one only.

        Optional. Will be simply skipped by :program:`mpm` if not implemented.
        """
        raise NotImplementedError

    def sync(self) -> None:
        """Refresh package metadata from remote repositories.

        Optional. Will be simply skipped by :program:`mpm` if not implemented.
        """
        raise NotImplementedError

    def cleanup(self) -> None:
        """Prune left-overs, remove orphaned dependencies and clear caches.

        Optional. Will be simply skipped by :program:`mpm` if not implemented.
        """
        raise NotImplementedError



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import ast
import inspect
import re
from pathlib import Path, PurePath
from string import ascii_letters, ascii_lowercase, digits

import pytest
from boltons.iterutils import unique
from boltons.urlutils import URL
from click_extra.platforms import ALL_PLATFORMS, Platform

from meta_package_manager.base import Operations, PackageManager
from meta_package_manager.cli import XKCD_MANAGER_ORDER
from meta_package_manager.pool import pool
from meta_package_manager.version import TokenizedString

from .conftest import all_managers, manager_classes

""" Test the structure, data and types returned by all package managers.

This test suite try to automate most of the basic reviewing work for the addition of
new package manager definitions.
"""


def test_xkcd_set():
    assert len(unique(XKCD_MANAGER_ORDER)) == len(XKCD_MANAGER_ORDER)
    assert set(pool.all_manager_ids).issuperset(XKCD_MANAGER_ORDER)


@all_managers
def test_deprecated(manager):
    if manager.deprecation_url is not None:
        location = URL(manager.deprecation_url)
        assert location
        assert location.scheme.lower() in ("http", "https")
        assert manager.deprecated is True


@all_managers
def test_ascii_id(manager):
    """All package manager IDs should be short ASCII strings."""
    assert manager.id
    assert manager.id.isascii()
    assert set(manager.id).issubset(ascii_lowercase + digits + "-")


@all_managers
def test_name(manager):
    """Check all managers have a name."""
    assert manager.name
    assert set(manager.name).issubset(ascii_letters + digits + "' ")


def test_unique_names():
    assert len({m.name for m in pool.values()}) == len(pool.all_manager_ids)


@all_managers
def test_homepage_url(manager):
    assert manager.homepage_url
    location = URL(manager.homepage_url)
    assert location
    assert location.scheme.lower() in ("http", "https")


@all_managers
def test_platforms(manager):
    """Check platforms is normalized as a frozenset."""
    assert manager.platforms
    assert isinstance(manager.platforms, frozenset)
    assert all(isinstance(p, Platform) for p in manager.platforms)
    assert ALL_PLATFORMS.issuperset(manager.platforms)


@all_managers
def test_requirement(manager):
    """Each manager is required to specify a minimal version or ``None``."""
    if manager.requirement is not None:
        assert set(manager.requirement).issubset(digits + ".")
        # Check provided string is lossless once passed via TokenizedString.
        assert str(TokenizedString(manager.requirement)) == manager.requirement


@all_managers
def test_cli_names_type(manager):
    """Check the pointed CLI name and path are file-system compatible."""
    assert manager.cli_names
    for name in manager.cli_names:
        assert name.isalnum()
        assert PurePath(name).name == name


@all_managers
def test_cli_search_path(manager):
    assert len(set(manager.cli_search_path)) == len(manager.cli_search_path)
    for search_path in manager.cli_search_path:
        path_obj = Path(search_path).resolve()
        assert path_obj.is_absolute()
        assert not path_obj.is_reserved()
        if path_obj.exists():
            assert path_obj.is_file() or path_obj.is_dir()


@all_managers
def test_extra_env_type(manager):
    """Check that definitions environment variables as a dict of strings."""
    if manager.extra_env:
        for key, value in manager.extra_env.items():
            for item in (key, value):
                assert item
                assert set(item).issubset(ascii_letters + digits + "_-")


@all_managers
def test_global_args_type(manager):
    """Check that definitions returns CLI args as a list of non-empty strings."""
    for global_args in (manager.pre_cmds, manager.pre_args, manager.post_args):
        assert all(global_args)
        for arg in global_args:
            assert set(arg).issubset(ascii_letters + digits + "-=+")


@all_managers
def test_version_cli_options(manager):
    """Version CLI options must be a list of non empty strings."""
    assert all(manager.version_cli_options)


@all_managers
def test_version_regex(manager):
    """Version regex is required.

    Check it compiles and match has a version group.
    """
    assert isinstance(manager.version_regex, str)
    regex = re.compile(manager.version_regex)
    assert "version" in regex.groupindex


@all_managers
def test_cli_path(manager):
    if manager.cli_path is not None:
        assert isinstance(manager.cli_path, Path)
        assert manager.cli_path.is_absolute()
        assert not manager.cli_path.is_reserved()
        assert manager.cli_path.is_file()


@pytest.mark.parametrize(
    ("query", "query_parts"),
    (
        ("cli-l-cli", {"cli", "l"}),
        ("ab12--cd34", {"ab12", "cd34"}),
        ("123/extra.34", {"123", "extra", "34"}),
        ("AB ab", {"AB", "ab"}),
    ),
)
def test_query_parts(query, query_parts):
    assert PackageManager.query_parts(query) == query_parts


def collect_props_ref():
    """Build the canonical reference from the base class.

    We need to parse the AST so we can collect both attributes and naked type
    annotations.
    """
    tree = ast.parse(Path(inspect.getfile(PackageManager)).read_bytes())

    manager_class = None
    for n in tree.body:
        if isinstance(n, ast.ClassDef) and n.name == "PackageManager":
            manager_class = n
            break

    for node in manager_class.body:
        if isinstance(node, ast.AnnAssign):
            yield node.target.id
        if isinstance(node, ast.Assign):
            yield from [t.id for t in node.targets]
        if isinstance(node, ast.FunctionDef):
            yield node.name


props_ref = tuple(collect_props_ref())


def test_operation_order():
    """Double check operation IDs are ordered and aligned to the base manager class and
    CLI implementation."""
    direct_operation_ids = [op for op in Operations.__members__ if op != "upgrade_all"]

    base_operations = [p for p in props_ref if p in direct_operation_ids]
    assert list(direct_operation_ids) == list(base_operations)

    cli_tree = ast.parse(Path(__file__).parent.joinpath("../cli.py").read_bytes())
    implemented_operations = [
        n.name
        for n in cli_tree.body
        if isinstance(n, ast.FunctionDef) and n.name in direct_operation_ids
    ]
    assert list(direct_operation_ids) == list(implemented_operations)


# Check the code of each file registered in the pool.
@manager_classes  # type: ignore[operator]
def test_content_order(manager_class):
    """Lint each package manager definition file to check its code structure is the same
    as the canonical PackageManager base class."""
    # Collect in order the IDs of all properties (ast.Assign) and functions in
    # the class.
    collected_props = []

    klass_tree = ast.parse(inspect.getsource(manager_class))
    for node in klass_tree.body:
        if isinstance(node, ast.Assign):
            collected_props.extend([t.id for t in node.targets])
        if isinstance(node, ast.FunctionDef):
            collected_props.append(node.name)

    enforced_props = tuple(p for p in collected_props if p in props_ref)
    expected_order = tuple(p for p in props_ref if p in collected_props)
    assert enforced_props == expected_order



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Utilities to generate extra labels to use for GitHub issues and PRs."""

from __future__ import annotations

from typing import Dict, FrozenSet

from boltons.iterutils import flatten

from .platforms import PLATFORM_GROUPS
from .pool import pool

LABELS: list[tuple[str, str, str]] = [
    (
        "🔌 bar-plugin",
        "#fef2c0",
        "Xbar/SwiftBar plugin code, documentation and features",
    ),
]
"""Global registry of all labels used in the project.

Structure:

.. code-block:: python

    ("label_name", "color", "optional_description")
"""


TLabelSet = FrozenSet[str]
TLabelGroup = Dict[str, TLabelSet]


def generate_labels(
    all_labels: TLabelSet,
    groups: TLabelGroup,
    prefix: str,
    color: str,
) -> dict[str, str]:
    """Generate labels."""
    # Check group definitions.
    group_entries = tuple(flatten(groups.values()))
    grouped_labels = set(group_entries)
    # Check there is no duplucates between groups.
    assert len(group_entries) == len(grouped_labels)
    # Check all labels to groups are referenced in the full label set.
    assert grouped_labels.issubset(all_labels)

    label_map = {}

    # Create a dedicated label for each non-grouped entry.
    standalone_labels = all_labels - grouped_labels
    for label_id in standalone_labels:
        label_name = f"{prefix}{label_id}"
        # Check the addition of the prefix does not collide with an existing label.
        assert label_name not in all_labels
        label_map[label_id] = label_name
        # Register label to the global registry.
        LABELS.append((label_name, color, label_id))

    # Create a dedicated label for each group.
    for group_id, label_ids in groups.items():
        label_name = f"{prefix}{group_id}"
        # Check the addition of the prefix does not collide with an existing label.
        assert label_name not in all_labels
        for label_id in label_ids:
            label_map[label_id] = label_name
        # Register label to the global registry.
        LABELS.append(
            (label_name, color, ", ".join(sorted(label_ids, key=str.casefold))),
        )

    # Sort label_map by their name.
    return dict(sorted(label_map.items(), key=lambda i: str.casefold(i[1])))


MANAGER_PREFIX = "📦 manager: "

MANAGER_LABEL_GROUPS: TLabelGroup = {
    "dnf-based": frozenset({"dnf", "yum"}),
    "dpkg-based": frozenset({"apt", "apt-mint", "opkg"}),
    "npm-based": frozenset({"npm", "yarn"}),
    "pacman-based": frozenset({"pacman", "pacaur", "paru", "yay"}),
    "pip-based": frozenset({"pip", "pipx"}),
}
"""Managers sharing some origin or implementation are grouped together under the same
label."""

all_manager_label_ids = frozenset(set(pool.all_manager_ids) | {"mpm"})
"""Adds ``mpm`` as its own manager alongside all those implemented."""

# Check group IDs do not collide with original labels.
assert all_manager_label_ids.isdisjoint(MANAGER_LABEL_GROUPS.keys())

MANAGER_LABELS = generate_labels(
    all_manager_label_ids,
    MANAGER_LABEL_GROUPS,
    MANAGER_PREFIX,
    "#bfdadc",
)
"""Maps all manager IDs to their labels."""


PLATFORM_PREFIX = "🖥 platform: "

PLATFORM_LABEL_GROUPS: TLabelGroup = {
    g.name: frozenset(p.name for p in g.platforms) for g in PLATFORM_GROUPS
}
"""Similar platforms are grouped together under the same label."""

all_platform_label_ids = frozenset(flatten(PLATFORM_LABEL_GROUPS.values()))

PLATFORM_LABELS = generate_labels(
    all_platform_label_ids,
    PLATFORM_LABEL_GROUPS,
    PLATFORM_PREFIX,
    "#bfd4f2",
)
"""Maps all platform names to their labels."""

# Force sorting of labels.
LABELS = sorted(LABELS, key=lambda i: str.casefold(i[0]))



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Registration, indexing and caching of package manager supported by ``mpm``."""

from __future__ import annotations

import logging
from functools import cached_property
from typing import TYPE_CHECKING, Final, Iterable, Iterator

from boltons.iterutils import unique
from click_extra.colorize import default_theme as theme

from .managers.apm import APM
from .managers.apt import APT, APT_Mint
from .managers.cargo import Cargo
from .managers.chocolatey import Choco
from .managers.composer import Composer
from .managers.dnf import DNF, YUM
from .managers.emerge import Emerge
from .managers.flatpak import Flatpak
from .managers.gem import Gem
from .managers.homebrew import Brew, Cask
from .managers.mas import MAS
from .managers.npm import NPM
from .managers.opkg import OPKG
from .managers.pacman import Pacaur, Pacman, Paru, Yay
from .managers.pip import Pip
from .managers.pipx import Pipx
from .managers.scoop import Scoop
from .managers.snap import Snap
from .managers.steamcmd import SteamCMD
from .managers.vscode import VSCode
from .managers.yarn import Yarn
from .managers.zypper import Zypper

if TYPE_CHECKING:
    from .base import Operations, PackageManager

manager_classes = (
    APM,
    APT,
    APT_Mint,
    Brew,
    Cargo,
    Cask,
    Choco,
    Composer,
    DNF,
    Emerge,
    Flatpak,
    Gem,
    MAS,
    NPM,
    OPKG,
    Pacaur,
    Pacman,
    Paru,
    Pip,
    Pipx,
    Scoop,
    Snap,
    SteamCMD,
    VSCode,
    YUM,
    Yarn,
    Yay,
    Zypper,
)
"""The list of all classes implementing the specific package managers.

Is considered valid package manager, definitions classes which:

#. are located in the :py:prop:`meta_package_manager.pool.ManagerPool.manager_subfolder`
    subfolder, and
#. are sub-classes of :py:class:`meta_package_manager.base.PackageManager`, and
#. are not :py:prop:`meta_package_manager.base.PackageManager.virtual` (i.e. have a
    non-null :py:prop:`meta_package_manager.base.PackageManager.cli_names` property).

These properties are checked and enforced in unittests.
"""


class ManagerPool:
    """A dict-like register, instanciating all supported package managers."""

    ALLOWED_EXTRA_OPTION: Final = frozenset(
        {"ignore_auto_updates", "stop_on_error", "dry_run", "timeout"},
    )
    """List of extra options that are allowed to be set on managers during the use of
    the :py:func:`meta_package_manager.pool.ManagerPool.select_managers` helper
    below."""

    @cached_property
    def register(self) -> dict[str, PackageManager]:
        """Instantiate all supported package managers."""
        register = {}
        for klass in manager_classes:
            manager = klass()
            register[manager.id] = manager
        return register

    # Emulates some dict methods.

    def __len__(self) -> int:
        return len(self.register)

    def __getitem__(self, key):
        return self.register[key]

    get = __getitem__

    def __iter__(self):
        yield from self.register

    def __contains__(self, key) -> bool:
        return key in self.register

    def values(self):
        return self.register.values()

    def items(self):
        return self.register.items()

    # Pre-compute all sorts of constants.

    @cached_property
    def all_manager_ids(self) -> tuple[str, ...]:
        """All recognized manager IDs.

        Returns a list of sorted items to provide consistency across all UI, and
        reproducibility in the order package managers are evaluated.
        """
        return tuple(sorted(self.register))

    @cached_property
    def maintained_manager_ids(self) -> tuple[str, ...]:
        """All manager IDs which are not deprecated."""
        return tuple(
            mid for mid in self.all_manager_ids if not self.register[mid].deprecated
        )

    @cached_property
    def default_manager_ids(self) -> tuple[str, ...]:
        """All manager IDs supported on the current platform and not deprecated.

        Must keep the same order defined by
        :py:prop:`meta_package_manager.pool.ManagerPool.all_manager_ids`.
        """
        return tuple(
            mid for mid in self.maintained_manager_ids if self.register[mid].supported
        )

    @cached_property
    def unsupported_manager_ids(self) -> tuple[str, ...]:
        """All manager IDs unsupported on the current platform but still maintained.

        Order is not important here as this list will be used to discard managers from
        selection sets.
        """
        return tuple(
            mid
            for mid in self.maintained_manager_ids
            if mid not in self.default_manager_ids
        )

    def select_managers(
        self,
        keep: Iterable[str] | None = None,
        drop: Iterable[str] | None = None,
        keep_deprecated: bool = False,
        keep_unsupported: bool = False,
        drop_inactive: bool = True,
        implements_operation: Operations | None = None,
        **extra_options: bool | int,
    ) -> Iterator[PackageManager]:
        """Utility method to extract a subset of the manager pool based on selection
        list (``keep`` parameter) and exclusion list (``drop`` parameter) criterion.

        By default, only the managers supported by the current platform are selected.
        Unless ``keep_unsupported`` is set to ``True``, in which case all managers
        implemented by ``mpm`` are selected, regardless of their supported platform.

        Deprecated managers are also excluded by default, unless ``keep_deprecated`` is
        ``True``.

        ``drop_inactive`` filters out managers that where not found on the system.

        ``implements_operation`` filters out managers which do not implements the
        provided operation.

        Finally, ``extra_options`` parameters are fed to manager objects to set some
        additional options.

        Returns a generator producing a manager instance one after the other.
        """
        # Produce the default set of managers to consider if none have been
        # provided by the ``keep`` parameter.
        if keep is None:
            if keep_deprecated:
                keep = self.all_manager_ids
            elif keep_unsupported:
                keep = self.maintained_manager_ids
            else:
                keep = self.default_manager_ids
        if drop is None:
            drop = set()
        assert set(self.all_manager_ids).issuperset(keep)
        assert set(self.all_manager_ids).issuperset(drop)

        assert self.ALLOWED_EXTRA_OPTION.issuperset(extra_options)

        # Reduce the set to the user's constraints.
        selected_ids = (mid for mid in unique(keep) if mid not in drop)

        # Deduplicate managers IDs while preserving order, then remove excluded
        # managers.
        for manager_id in selected_ids:
            manager = self.register[manager_id]

            # Check if operation is not implemented before calling `.available`. It
            # saves one call to the package manager CLI.
            if implements_operation and not manager.implements(implements_operation):
                logging.warning(
                    f"{theme.invoked_command(manager_id)} "
                    f"does not implement {implements_operation}.",
                )
                continue

            # Filters out inactive managers.
            if drop_inactive and not manager.available:
                logging.warning(
                    f"Skip unavailable {theme.invoked_command(manager_id)} manager.",
                )
                continue

            # Apply manager-level options.
            for param, value in extra_options.items():
                assert hasattr(manager, param)
                setattr(manager, param, value)

            yield manager


pool = ManagerPool()



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

from dataclasses import asdict

import pytest
from boltons.iterutils import flatten

from meta_package_manager.pool import pool
from meta_package_manager.specifier import PURL_MAP, EmptyReduction, Solver, Specifier


def test_purl_map():
    # Check all our supported managers are registered.
    assert set(pool.all_manager_ids).issubset(PURL_MAP)
    # Check our hard-coded purl mapping points to implemented managers.
    assert (
        set(pool.all_manager_ids)
        .union({None})
        .issuperset(set(flatten(PURL_MAP.values())))
    )


def props(spec: Specifier):
    """Utility to help compares specifiers between themselves.

    I.e. all properties of specifiers are the same but the ``raw_specs`` property.
    """
    return sorted((k, v) for k, v in asdict(spec).items() if k != "raw_spec")


@pytest.mark.parametrize(
    ("spec_string", "expected"),
    (
        pytest.param(
            "leftpad",
            ({"package_id": "leftpad", "manager_id": None, "version": None},),
            id="ascii",
        ),
        pytest.param(
            "left_pad",
            ({"package_id": "left_pad", "manager_id": None, "version": None},),
            id="underscore",
        ),
        pytest.param(
            "left.pad",
            ({"package_id": "left.pad", "manager_id": None, "version": None},),
            id="dot",
        ),
        pytest.param(
            "left-pad",
            ({"package_id": "left-pad", "manager_id": None, "version": None},),
            id="dash",
        ),
        pytest.param(
            "left-pad@8.a7",
            ({"package_id": "left-pad", "manager_id": None, "version": "8.a7"},),
            id="version_spec",
        ),
        pytest.param(
            "pkg:npm/leftpad",
            (
                {
                    "package_id": "leftpad",
                    "manager_id": "npm",
                    "version": None,
                },
            ),
            id="purl_with_manager",
        ),
        pytest.param(
            "pkg:npm/leftpad@1.2.3",
            ({"package_id": "leftpad", "manager_id": "npm", "version": "1.2.3"},),
            id="purl_with_version",
        ),
        pytest.param(
            "pkg:rubygems/dummy@21.0-b",
            ({"package_id": "dummy", "manager_id": "gem", "version": "21.0-b"},),
            id="purl_single_alias",
        ),
        pytest.param(
            "pkg:rpm/ping@2011-04.gamma",
            (
                {
                    "package_id": "ping",
                    "manager_id": "dnf",
                    "version": "2011-04.gamma",
                },
                {
                    "package_id": "ping",
                    "manager_id": "yum",
                    "version": "2011-04.gamma",
                },
                {
                    "package_id": "ping",
                    "manager_id": "zypper",
                    "version": "2011-04.gamma",
                },
            ),
            id="purl_multiple_aliases",
        ),
    ),
)
def test_parse_specs(spec_string, expected):
    specs = tuple(Specifier.from_string(spec_string))

    spec_props = []
    for spec in specs:
        assert spec.raw_spec == spec_string
        # Serialize each specifier into an hashable set without the 'raw_spec'
        # field (which we checked above).
        spec_props.append(props(spec))

    assert sorted(spec_props) == sorted(sorted(s.items()) for s in expected)


@pytest.mark.parametrize(
    ("spec_strings", "target_managers", "expected"),
    (
        pytest.param(
            ("leftpad",),
            None,
            Specifier.from_string("leftpad"),
            id="tuple_specs",
        ),
        pytest.param(
            ["leftpad"],
            None,
            Specifier.from_string("leftpad"),
            id="list_specs",
        ),
        pytest.param(
            {"leftpad"},
            None,
            Specifier.from_string("leftpad"),
            id="set_specs",
        ),
        pytest.param(
            ("leftpad", "leftpad"),
            None,
            Specifier.from_string("leftpad"),
            id="duplicates",
        ),
        pytest.param(
            {"left-pad@8.a7"},
            None,
            Specifier.from_string("left-pad@8.a7"),
            id="version_spec",
        ),
        pytest.param(
            {"left-pad", "left-pad@8.23.a7"},
            None,
            Specifier.from_string("left-pad@8.23.a7"),
            id="version_and_unversionned_mix",
        ),
        pytest.param(
            {"pkg:npm/leftpad"},
            None,
            Specifier.from_string("pkg:npm/leftpad"),
            id="purl_with_manager",
        ),
        pytest.param(
            {"pkg:npm/leftpad@1.2.3"},
            None,
            Specifier.from_string("pkg:npm/leftpad@1.2.3"),
            id="purl_with_version",
        ),
        pytest.param(
            {"pkg:rubygems/dummy@21.0-b"},
            None,
            Specifier.from_string("pkg:gem/dummy@21.0-b"),
            id="purl_single_alias",
        ),
        pytest.param(
            {"pkg:rpm/ping@2011-04.gamma"},
            ["dnf", "yum", "zypper"],
            Specifier.from_string("pkg:dnf/ping@2011-04.gamma"),
            id="purl_multiple_aliases_priority1",
        ),
        pytest.param(
            {"pkg:rpm/ping@2011-04.gamma"},
            ["zypper", "yum"],
            Specifier.from_string("pkg:zypper/ping@2011-04.gamma"),
            id="purl_multiple_aliases_priority2",
        ),
        pytest.param(
            {"pkg:rpm/ping@2011-04.gamma"},
            ["pypi"],
            EmptyReduction,
            id="purl_multiple_aliases_unmatching_manager_skip_package",
        ),
        pytest.param(
            {"pkg:rpm/ping@2011-04.gamma"},
            None,
            ValueError,
            id="unresolveable_multiple_aliases",
        ),
        pytest.param(
            {"pkg:npm/leftpad", "leftpad@77.10.0"},
            None,
            Specifier.from_string("pkg:npm/leftpad"),
            id="purl_takes_precedence",
        ),
        pytest.param(
            {"leftpad", "leftpad@1.7.3", "pkg:npm/leftpad@1.7.3"},
            None,
            Specifier.from_string("pkg:npm/leftpad@1.7.3"),
            id="mixed_package_specs",
        ),
        pytest.param(
            {"left-pad@33.1.a", "left-pad@0100"},
            None,
            Specifier.from_string("left-pad@0100"),
            id="multiple_versions",
        ),
        pytest.param(
            {"left-pad@99.00", "left-pad@99.00"},
            None,
            Specifier.from_string("left-pad@99.00"),
            id="duplicate_version_spec",
        ),
        pytest.param(
            {"left-pad@99.00", "left-pad@99"},
            None,
            Specifier.from_string("left-pad@99.00"),
            id="similar_version_spec",
        ),
        pytest.param(
            {"left-pad@99.00", "left-pad@0099.00"},
            None,
            ValueError,
            id="equivalent_version_spec",
        ),
    ),
)
def test_reduce_specs(spec_strings, target_managers, expected):
    solver = Solver(spec_strings, manager_priority=target_managers)

    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            solver.reduce_specs(solver.spec_pool)

    else:
        reduced_spec = solver.reduce_specs(solver.spec_pool)
        expected = tuple(expected)
        assert len(expected) == 1
        assert props(reduced_spec) == props(expected[0])



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re
import subprocess
from collections import Counter
from itertools import product

import pytest
from boltons.iterutils import flatten
from click_extra.testing import args_cleanup, env_copy
from click_extra.tests.conftest import unless_macos

from meta_package_manager import bar_plugin
from meta_package_manager.version import parse_version


def _invokation_matrix(*iterables):
    """Pre-compute a matrix of all possible options for invokation."""
    for args in product(*iterables):
        yield args_cleanup(args)


def _shell_invokation_matrix():
    """Pre-compute a matrix of all possible options used for shell invokation.

    See the list of shell supported by SwiftBar at:
    https://github.com/swiftbar/SwiftBar/commit/366695d594884fe141bc1752ab0f25d2c43334fa

    Returns
    -------
    .. code-block:: python
        (
            ("bash", "-c"),
            ("bash", "--login", "-c"),
            ("/bin/bash", "-c"),
            ("/bin/bash", "--login", "-c"),
            ("zsh", "-c"),
            ("zsh", "--login", "-c"),
            ("/bin/zsh", "-c"),
            ("/bin/zsh", "--login", "-c"),
            ("/usr/bin/env", "bash", "-c"),
            ("/usr/bin/env", "bash", "--login", "-c"),
            ("/usr/bin/env", "/bin/bash", "-c"),
            ("/usr/bin/env", "/bin/bash", "--login", "-c"),
            ("/usr/bin/env", "zsh", "-c"),
            ("/usr/bin/env", "zsh", "--login", "-c"),
            ("/usr/bin/env", "/bin/zsh", "-c"),
            ("/usr/bin/env", "/bin/zsh", "--login", "-c"),
        )
    """
    return _invokation_matrix(
        # Env prefixes.
        (None, "/usr/bin/env"),
        # Naked and full binary paths.
        flatten((bin_id, f"/bin/{bin_id}") for bin_id in ("bash", "zsh")),
        # Options.
        (
            "-c",
            # XXX Login shell defaults to Python 2.7 on GitHub macOS runners and is
            # picked up by surprise for bar plugin tests:
            # Traceback (most recent call last):
            #   File "/Library/Frameworks/.../2.7/lib/python2.7/runpy.py",
            #   line 163, in _run_module_as_main
            #     mod_name, _Error)
            #   File "/Library/Frameworks/.../2.7/lib/python2.7/runpy.py",
            #   line 111, in _get_module_details
            #     __import__(mod_name)  # Do not catch exceptions initializing package
            #   File "meta_package_manager/__init__.py", line 33, in <module>
            #     from click_extra.logging import logger
            # ImportError: No module named click_extra.logging
            # ("--login", "-c"),
        ),
    )


def _python_invokation_matrix():
    """Pre-compute a matrix of all possible options used for python invokation.

    Returns
    -------
    .. code-block:: python
        (
            ("python",),
            ("python3",),
            ("/usr/bin/env", "python"),
            ("/usr/bin/env", "python3"),
        )
    """
    return _invokation_matrix(
        # Env prefixes.
        (None, "/usr/bin/env"),
        # Binary paths
        ("python", "python3"),
    )


def _subcmd_args(invoke_args: tuple[str, ...] | None, *subcmd_args: str):
    """Cleanup args and eventually concatenate all ``subcmd_args`` items to a space
    separated string if ``invoke_args`` is defined and its last argument is equal to
    ``-c``."""
    raw_args: list[str] = []
    if invoke_args:
        raw_args.extend(invoke_args)
        if invoke_args[-1] == "-c":
            subcmd_args = (" ".join(subcmd_args),)
    raw_args.extend(subcmd_args)
    return args_cleanup(raw_args)


@unless_macos
class TestBarPlugin:
    common_checklist = [
        # Menubar line. Required.
        (r"(🎁↑\d+|📦✓)( ⚠️\d+)? \| dropdown=false$", True),
        # Submenus and sections marker. Required.
        (r"-{3,5}$", True),
        # Upgrade all line. Required.
        (
            r"(--)?🆙 Upgrade all \S+ packages? \| shell=\S+( param\d+=\S+)+ "
            r"refresh=true terminal=(false|true alternate=true)$",
            True,
        ),
        # Error line. Optional.
        (
            r"(--)?.+ \| font=Menlo size=12 color=red trim=false "
            r"ansi=false emojize=false( symbolize=false)?$",
            False,
        ),
    ]

    def plugin_output_checks(self, checklist, extra_env=None):
        """Run the plugin script and check its output against the checklist."""
        process = subprocess.run(
            # Force the plugin to be called within Poetry venv to not have it seeking
            # for macOS's default Python.
            ("poetry", "run", "python", bar_plugin.__file__),
            capture_output=True,
            encoding="utf-8",
            env=env_copy(extra_env),
        )

        assert process.returncode == 0
        assert not process.stderr

        checks = checklist + self.common_checklist

        match_counter = Counter()

        for line in process.stdout.splitlines():
            # The line is expected to match at least one regex.
            matches = False
            for index, (regex, _) in enumerate(checks):
                if re.match(regex, line):
                    matches = True
                    match_counter[index] += 1
                    break
            if not matches:
                print(process.stdout)
                msg = f"plugin output line {line!r} did not match any regex."
                raise Exception(msg)

        # Check all required regex did match at least once.
        for index, (regex, required) in enumerate(checks):
            if required and not match_counter[index]:
                msg = f"{regex!r} regex did not match any plugin output line."
                raise Exception(msg)

    @pytest.mark.xdist_group(name="avoid_concurrent_plugin_runs")
    @pytest.mark.parametrize("submenu_layout", (True, False, None))
    @pytest.mark.parametrize("table_rendering", (True, False, None))
    def test_rendering(self, submenu_layout, table_rendering):
        extra_checks = []
        if table_rendering is False:
            extra_checks.extend(
                (
                    # Package manager section header.
                    (r"(⚠️ )?\d+ outdated .+ packages?", True),
                    # Package upgrade line.
                    (
                        r"(--)?\S+ \S+ → \S+ \| shell=\S+( param\d+=\S+)+ "
                        r"refresh=true terminal=(false|true alternate=true)$",
                        True,
                    ),
                ),
            )
        # Default case is VAR_TABLE_RENDERING=true.
        else:
            extra_checks.extend(
                (
                    # Package manager section header.
                    (r"(⚠️ )?\S+ - \d+ packages?\s+\| font=Menlo size=12", True),
                    # Package upgrade line.
                    (
                        r"(--)?\S+\s+\S+ → \S+\s+\| shell=\S+( param\d+=\S+)+ "
                        r"font=Menlo size=12 refresh=true "
                        r"terminal=(false|true alternate=true)?$",
                        True,
                    ),
                ),
            )

        extra_env = {}
        if submenu_layout is not None:
            extra_env["VAR_SUBMENU_LAYOUT"] = str(submenu_layout)
        if table_rendering is not None:
            extra_env["VAR_TABLE_RENDERING"] = str(table_rendering)

        self.plugin_output_checks(extra_checks, extra_env=extra_env)

    @pytest.mark.xdist_group(name="avoid_concurrent_plugin_runs")
    @pytest.mark.parametrize("shell_args", (None, *_shell_invokation_matrix()))
    def test_plugin_shell_invokation(self, shell_args):
        """Test execution of plugin on different shells.

        Do not execute the complete search for outdated packages, just stop at searching
        for the mpm executable and extract its version.
        """
        process = subprocess.run(
            _subcmd_args(shell_args, bar_plugin.__file__, "--check-mpm"),
            capture_output=True,
            encoding="utf-8",
        )

        assert process.returncode == 0
        assert not process.stderr
        assert re.match(r"^.+ v\d+\.\d+\.\d+$", process.stdout)

    @pytest.mark.parametrize("shell_args", (None, *_shell_invokation_matrix()))
    @pytest.mark.parametrize("python_args", _python_invokation_matrix())
    def test_python_shell_invokation(self, shell_args, python_args):
        """Test any Python shell invokation is properly configured and all are
        compatible with plugin requirements."""
        process = subprocess.run(
            _subcmd_args(shell_args, *python_args, "--version"),
            capture_output=True,
            encoding="utf-8",
        )

        assert process.returncode == 0
        assert process.stdout
        assert not process.stderr

        # We need to parse the version to account for alpha release,
        # like Python `3.12.0a4`.
        assert parse_version(process.stdout.split()[1]) >= parse_version(
            ".".join(str(i) for i in bar_plugin.python_min_version),
        )



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Homebrew-specific tests."""

from __future__ import annotations

import subprocess
from pathlib import Path

import pytest
from click_extra.testing import env_copy
from click_extra.tests.conftest import unless_macos


@pytest.fixture()
def install_cask():
    """A fixture to install a Cask from a specific commit."""

    packages = set()
    """List of installed packages."""

    def get_cask_tap_path():
        """Default location of Homebrew Cask formulas on macOS.

        This is supposed to be a `shallow copy of the official cask repository
        <https://github.com/Homebrew/homebrew-cask>`_.

        The later can be obtained with:

        .. code-block:: shell-session
            $ export HOMEBREW_NO_INSTALL_FROM_API="1"
            $ brew tap homebrew/cask
        """
        process = subprocess.run(
            ("brew", "--repository"),
            capture_output=True,
            encoding="utf-8",
        )
        assert process.returncode == 0
        assert not process.stderr
        assert process.stdout

        brew_root = Path(process.stdout.strip())
        assert brew_root.is_absolute()
        assert brew_root.exists()
        assert brew_root.is_dir()

        cask_repo = brew_root.joinpath("Library/Taps/homebrew/homebrew-cask/Casks/")
        assert cask_repo.is_absolute()
        assert cask_repo.exists()
        assert cask_repo.is_dir()
        return cask_repo

    def get_cask_folder(package_id: str):
        """Get the folder where a Cask is located.

        ..warning::
            As of `aa46114 <https://github.com/Homebrew/homebrew-cask/commit/aa46114>`_,
            casks have been moved to a subfolder named after their first letter.
        """
        cask_folder = get_cask_tap_path().joinpath(package_id[0])
        assert cask_folder.is_absolute()
        assert cask_folder.exists()
        assert cask_folder.is_dir()
        return cask_folder

    def git_checkout(package_id: str, commit: str):
        process = subprocess.run(
            (
                "git",
                "-C",
                get_cask_folder(package_id),
                "checkout",
                commit,
                f"{package_id}.rb",
            ),
        )
        assert not process.stderr
        assert process.returncode == 0

    def brew_cleanup():
        """Remove all downloads cached by Homebrew.

        .. note::
            ``brew`` does not cleanup ``~/Library/Caches/Homebrew``, see:
            https://github.com/Homebrew/brew/issues/3784#issuecomment-364675767

            We might need to force it with:

            .. code-block:: shell-session
                $ rm -rf $(brew --cache)
        """
        process = subprocess.run(
            ("brew", "cleanup", "-s", "--prune=all"),
            env=env_copy({"HOMEBREW_NO_AUTO_UPDATE": "1"}),
        )
        assert not process.stderr
        assert process.returncode == 0

    def brew_uninstall(package_id):
        process = subprocess.run(("brew", "uninstall", "--cask", "--force", package_id))
        assert not process.stderr
        assert process.returncode == 0

    def _install_cask(package_id, commit):
        packages.add(package_id)

        # Fetch locally the old version of the Cask's formula.
        git_checkout(package_id, commit)
        brew_cleanup()

        # XXX Work around the bad old ubersicht cask formula.
        # Replace ":yosemite" by ":el_capitan" like the last commit did at:
        # https://github.com/Homebrew/homebrew-cask/commit/5cddbd6
        # The day we have a new version of ubersicht after this commit, we will be able
        # to remove that hack, and have the install_cask() invokation from the
        # test_autoupdate_unicode_name() test below checkout ubersicht from that commit.
        if package_id == "ubersicht":
            cask_path = get_cask_tap_path().joinpath(f"{package_id}.rb")
            content = cask_path.read_text()
            cask_path.write_text(content.replace(":yosemite", ":el_capitan"))

        # Install the cask but bypass its local cache auto-update: we want to
        # force brew to rely on our formula from the past.
        process = subprocess.run(
            ("brew", "reinstall", "--cask", package_id),
            capture_output=True,
            encoding="utf-8",
            check=True,
            env=env_copy(
                {
                    # Do not let brew use its live API to fetch the latest version.
                    # This variable forces brew to use the local repository instead:
                    # https://github.com/Homebrew/brew/blob/10845a1/Library/Homebrew/env_config.rb#L314-L317
                    "HOMEBREW_NO_INSTALL_FROM_API": "1",
                    "HOMEBREW_NO_AUTO_UPDATE": "1",
                    "HOMEBREW_NO_ENV_HINTS": "1",
                },
            ),
        )

        # Restore old formula to its most recent version.
        git_checkout(package_id, "HEAD")
        brew_cleanup()

        # Check the cask has been properly installed.
        if process.stderr:
            assert "is already installed" not in process.stderr
        assert f"{package_id} was successfully installed!" in process.stdout
        return process.stdout

    yield _install_cask

    # Remove all installed packages.
    for package_id in packages:
        brew_uninstall(package_id)


@pytest.mark.destructive()
@unless_macos
# XXX The layout of the cask repository has changed, so we cannot go back in time too
# much and checkout old versions of casks. We need to wait for a couple of new releases
# of the casks we use in our tests, so we can have enough commit depth to test the
# upgrade process.
@pytest.mark.skip(
    reason="Cask repository structure changed too much to go back in time."
)
class TestCask:
    @pytest.mark.xdist_group(name="avoid_concurrent_git_tweaks")
    def test_autoupdate_unicode_name(self, invoke, install_cask):
        """See #16."""
        # Install an old version of a package with a unicode name.
        # Old Cask formula for ubersicht 1.6.69.
        # XXX Update this commit when a new version of ubersicht is released so we can
        # get rid of the hack above.
        output = install_cask("ubersicht", "9870e8ffaa8dc83403973580415b2c56dc760f15")
        assert "Uebersicht-1.6.69.app.zip" in output
        assert "Übersicht.app" in output
        assert "Übersicht.app" not in output

        # Ubersicht is not reported as outdated because is tagged as
        # auto-update.
        result = invoke("--cask", "outdated")
        assert result.exit_code == 0
        assert "ubersicht" not in result.stdout
        assert "Übersicht" not in result.stdout

        # Try with explicit option.
        result = invoke("--ignore-auto-updates", "--cask", "outdated")
        assert result.exit_code == 0
        assert "ubersicht" not in result.stdout
        assert "Übersicht" not in result.stdout

        # Look for reported available upgrade.
        result = invoke("--include-auto-updates", "--cask", "outdated")
        assert result.exit_code == 0
        assert "ubersicht" in result.stdout
        # Outdated subcommand does not fetch the unicode name by default.
        assert "Übersicht" not in result.stdout

    @pytest.mark.xdist_group(name="avoid_concurrent_git_tweaks")
    def test_multiple_names(self, invoke, install_cask):
        """See #26."""
        # Install an old version of a package with multiple names.
        # Old Cask formula for xld 20210101,153.1.
        output = install_cask("xld", "6be43816ea729c18219322f656c62a702e1c2440")
        assert "xld-20210101.dmg" in output
        assert "XLD.app" in output

        # Look for reported available upgrade.
        result = invoke("--include-auto-updates", "--cask", "outdated")
        assert result.exit_code == 0
        assert "xld" in result.stdout
        # Outdated subcommand does not fetch the unicode name by default.
        assert "X Lossless Decoder" not in result.stdout



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


from __future__ import annotations

import dataclasses
import json
import re
from time import sleep

import pytest
from boltons.iterutils import same
from click_extra.tests.conftest import skip_windows, unless_macos

from meta_package_manager.base import Operations, Package
from meta_package_manager.pool import pool

from .test_cli import CLISubCommandTests, CLITableTests


@pytest.fixture()
def subcmd():
    return "search", "abc"


class TestSearch(CLISubCommandTests, CLITableTests):
    @staticmethod
    def evaluate_signals(mid, stdout, stderr):
        yield from (
            f"warning: {mid} does not implement {Operations.search}." in stderr,
            # Common "not found" warning message.
            f"warning: Skip unavailable {mid} manager." in stderr,
            # Stats line at the end of output.
            f"{mid}: " in stderr.splitlines()[-1] if stderr else "",
        )

    def test_json_parsing(self, invoke, subcmd):
        result = invoke("--output-format", "json", subcmd)
        assert result.exit_code == 0
        data = json.loads(result.stdout)

        assert data
        assert isinstance(data, dict)
        assert set(data).issubset(pool.default_manager_ids)

        for manager_id, info in data.items():
            assert isinstance(manager_id, str)
            assert isinstance(info, dict)

            assert isinstance(info["id"], str)
            assert isinstance(info["name"], str)

            assert set(info) == {"errors", "id", "name", "packages"}

            assert isinstance(info["errors"], list)
            if info["errors"]:
                assert same(map(type, info["errors"]), str)

            assert info["id"] == manager_id

            assert isinstance(info["packages"], list)
            for pkg in info["packages"]:
                assert isinstance(pkg, dict)

                fields = {f.name for f in dataclasses.fields(Package)}
                assert set(pkg).issubset(fields)

                for f in pkg:
                    assert isinstance(pkg[f], str) or pkg[f] is None

    @unless_macos
    def test_unicode_search(self, invoke):
        """See #16."""
        result = invoke("--mas", "search", "钉钉")
        assert result.exit_code == 0
        assert "钉钉" in result.stdout

    # PyPi's online search API was at first rate-limited. So we added an artificial
    # 2-seconds delay to prevent the following error:
    #   xmlrpc.client.Fault: <Fault -32500: 'HTTPTooManyRequests: The action could not
    #   be performed because there were too many requests by the client. Limit may
    #   reset in 1 seconds.'>
    # Then the search API was shutdown altogether as it was hammered (see
    # https://github.com/pypa/pip/issues/5216#issuecomment-744605466) which produced
    # this error:
    #   xmlrpc.client.Fault: <Fault -32500: "RuntimeError: PyPI's XMLRPC API has been
    #   temporarily disabled due to unmanageable load and will be deprecated in the
    #   near future. See https://status.python.org/ for more information.">

    skip_pip_search = pytest.mark.skip(reason="pip search is deprecated")

    @skip_pip_search
    def test_exact_search_one_result(self, invoke):
        sleep(2)
        result = invoke("--pip", "search", "--exact", "sed", color=False)
        assert result.exit_code == 0
        assert "1 package total" in result.stdout
        assert " sed " in result.stdout

    @skip_pip_search
    @pytest.mark.parametrize(
        "query",
        ("SED", "SeD", "sEd*", "*sED*", "_seD-@", "", "_"),
    )
    def test_exact_search_no_result(self, invoke, query):
        sleep(2)
        result = invoke("--pip", "search", "--exact", query)
        assert result.exit_code == 0
        assert "0 package total" in result.stdout
        assert "sed" not in result.stdout

    @skip_pip_search
    @pytest.mark.parametrize("query", ("", "_", "_seD-@"))
    def test_fuzzy_search_no_results(self, invoke, query):
        sleep(2)
        result = invoke("--pip", "search", query)
        assert result.exit_code == 0
        assert "0 package total" in result.stdout
        assert "sed" not in result.stdout

    @skip_pip_search
    @pytest.mark.parametrize("query", ("sed", "SED", "SeD", "sEd*", "*sED*"))
    def test_fuzzy_search_multiple_results(self, invoke, query):
        sleep(2)
        result = invoke("--pip", "search", query, color=False)
        assert result.exit_code == 0
        assert "2 packages total" in result.stdout
        assert " sed " in result.stdout
        assert " SED-cli " in result.stdout

    @skip_pip_search
    @pytest.mark.parametrize("query", ("", "_", "_seD-@"))
    def test_extended_search_no_results(self, invoke, query):
        sleep(2)
        result = invoke("--pip", "search", "--extended", query)
        assert result.exit_code == 0
        assert "0 package total" in result.stdout
        assert "sed" not in result.stdout

    @skip_pip_search
    @pytest.mark.parametrize("query", ("sed", "SED", "SeD", "sEd*", "*sED*"))
    def test_extended_search_multiple_results(self, invoke, query):
        sleep(2)
        result = invoke("--pip", "search", "--extended", query)
        assert result.exit_code == 0
        last_line = result.stdout.splitlines()[-1]
        assert last_line
        msg_match = re.match(
            r"^([0-9]+) packages? total \(pip: ([0-9]+)\).$",
            last_line,
        )
        assert msg_match
        assert same(msg_match.groups())
        # We should find lots of results for this package search.
        assert int(msg_match.groups()[0]) >= 20

    # XXX Colorama, a dependency of Click, is supposed to transform ANSI codes into
    # Windows-specific directives but doen't. Maybe because the running environment is
    # not considered as a terminal. We need to investigate this further and report
    # issues to either Click or Colorama.
    @skip_windows
    def test_search_highlight(self, invoke):
        """We search on cargo as it is available on all platforms.

        The search on the small ``co`` string is producing all variations of multiple
        ``co`` substring matches and highlights.
        """
        result = invoke("--cargo", "search", "co")
        assert result.exit_code == 0
        # co
        assert "│ \x1b[32m\x1b[1mco\x1b[0m " in result.stdout
        # acco
        assert "│ ac\x1b[32m\x1b[1mco\x1b[0m " in result.stdout
        # bicoro
        assert "│ bi\x1b[32m\x1b[1mco\x1b[0mro " in result.stdout
        # cocomo-core
        assert (
            "│ \x1b[32m\x1b[1mcoco\x1b[0mmo-\x1b[32m\x1b[1mco\x1b[0mre "
            in result.stdout
        )
        # cocomo-tui
        assert "│ \x1b[32m\x1b[1mcoco\x1b[0mmo-tui " in result.stdout



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import pytest

from meta_package_manager.pool import pool

from .conftest import default_manager_ids
from .test_cli import CLISubCommandTests


@pytest.fixture()
def subcmd(create_config):
    """Seed common subcommand tests with a dummy file and content to allow the CLI to
    not fail on required file input."""
    toml_path = create_config(
        "dummy.toml",
        """
        [dummy_manager]
        fancy_package = "0.0.1"
        """,
    )
    return "restore", str(toml_path)


@pytest.mark.skip(reason="Generated config file is not isolated from other tests.")
class TestRestore(CLISubCommandTests):
    @staticmethod
    def evaluate_signals(mid, stdout, stderr):
        yield from (
            # Warning message for restore command.
            f"warning: No [{mid}] section found." in stderr,
            # Restoring message.
            f"Restore {mid} packages..." in stderr,
            # Common "not found" warning message.
            f"warning: Skip unavailable {mid} manager." in stderr,
        )

    @pytest.mark.destructive()
    def test_default_all_managers(self, invoke, create_config):
        toml_path = create_config(
            "all-managers.toml",
            "".join(
                f"""
            [{m}]
            blah = 123
            """
                for m in pool.all_manager_ids
            ),
        )

        result = invoke("--verbosity", "INFO", "restore", str(toml_path))
        assert result.exit_code == 0
        assert "all-managers.toml" in result.stderr
        self.check_manager_selection(result)

    @pytest.mark.destructive()
    @default_manager_ids
    def test_single_manager(self, invoke, create_config, manager_id):
        toml_path = create_config(
            "all-managers.toml",
            "".join(
                f"""
            [{m}]
            blah = 123
            """
                for m in pool.all_manager_ids
            ),
        )

        result = invoke(
            "--verbosity", "INFO", f"--{manager_id}", "restore", str(toml_path)
        )
        assert result.exit_code == 0
        self.check_manager_selection(result, {manager_id})

    def test_ignore_unrecognized_manager(self, invoke, create_config):
        toml_path = create_config(
            "unrecognized.toml",
            """
            [random_section]
            blah = 123
            """,
        )

        result = invoke("--verbosity", "INFO", "restore", str(toml_path))
        assert result.exit_code == 0
        assert "unrecognized.toml" in result.stderr
        assert "Ignore [random_section] section" in result.stderr

    @pytest.mark.destructive()
    def test_restore_single_manager(self, invoke, create_config):
        toml_path = create_config(
            "pip-npm-dummy.toml",
            """
            [pip]
            leftpad = "0.1.2"

            [npm]
            chance = "1.1.9"
            """,
        )

        result = invoke(
            "--verbosity", "INFO", "--npm", "restore", str(toml_path), color=False
        )
        assert result.exit_code == 0
        assert "pip-npm-dummy.toml" in result.stderr
        assert "Restore pip packages..." not in result.stderr
        assert "Restore npm packages..." in result.stderr

    @pytest.mark.destructive()
    def test_restore_excluded_manager(self, invoke, create_config):
        toml_path = create_config(
            "pip-npm-dummy.toml",
            """
            [pip]
            leftpad = "0.1.2"

            [npm]
            chance = "1.1.9"
            """,
        )

        result = invoke(
            "--verbosity",
            "INFO",
            "--exclude",
            "npm",
            "restore",
            str(toml_path),
            color=False,
        )
        assert result.exit_code == 0
        assert "pip-npm-dummy.toml" in result.stderr
        assert "Restore pip packages..." in result.stderr
        assert "Restore npm packages..." not in result.stderr

    def test_empty_manager(self, invoke, create_config):
        toml_path = create_config(
            "pip-empty.toml",
            """
            [pip]
            """,
        )

        result = invoke("--verbosity", "INFO", "restore", str(toml_path), color=False)
        assert result.exit_code == 0
        assert "pip-empty.toml" in result.stderr
        assert "Restore pip packages..." in result.stderr



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Utilities to manage and resolve constraints from a set of package specifiers."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from functools import cached_property
from itertools import groupby
from operator import attrgetter
from typing import Final, Iterable, Iterator, Sequence

from packageurl import PackageURL

from .pool import pool
from .version import TokenizedString, parse_version

VERSION_SEP: Final = "@"
"""Separator used by ``mpm`` to split package's ID from its version:

This has been chosen as a separator because it is shared by popular package managers
and purls.

..code-block::

    package_id@version
"""


# 1:1 mapping between purl types and mpm manager IDs.
PURL_MAP: dict[str, set[str] | None] = {mid: {mid} for mid in pool.all_manager_ids}
# Manager IDs collected by looking at packageurl-python source code.
PURL_MAP.update(
    {
        "alpine": None,
        "bitbucket": None,
        "deb": {"apt", "apt-mint"},
        "docker": None,
        "generic": None,
        "github": None,
        "gitlab": None,
        "golang": None,
        "hackage": None,
        "maven": None,
        "nuget": None,
        "pypi": {"pip", "pipx"},
        "rpm": {"dnf", "yum", "zypper"},
        "rubygems": {"gem"},
        "sourceforge": None,
    },
)


@dataclass(frozen=True)
class Specifier:
    """Lightweight representation of a package specification.

    Contains all parsed metadata to be used as constraints.
    """

    raw_spec: str
    """Original, un-parsed specifier string provided by the user."""

    package_id: str
    """ID is required and is the primary key used for specification."""

    manager_id: str | None = None

    version: str | None = None
    """Version string, a 1:1 copy of the one provided by the user."""

    @classmethod
    def parse_purl(cls, spec_str: str) -> tuple[Specifier, ...] | None:
        """Parse a purl string.

        Yields ``Specifier`` objects or returns ``None``.
        """
        # Try to parse specifier as a purl.
        try:
            purl = PackageURL.from_string(spec_str)
        except ValueError as ex:
            logging.debug(f"{spec_str} is not a purl: {ex}")
            return None

        # Specifier is a purl, extract its metadata.
        manager_ids = PURL_MAP.get(purl.type)
        if not manager_ids:
            msg = f"Unrecognized {purl.type} purl type."
            raise ValueError(msg)

        # The purl can be handled by one manager or more.
        return tuple(
            Specifier(
                raw_spec=spec_str,
                package_id=purl.name,
                manager_id=manager_id,
                version=purl.version,
            )
            for manager_id in manager_ids
        )

    @classmethod
    def from_string(cls, spec_str: str) -> Iterator[Specifier]:
        """Parse a string into a package specifier.

        Supports various formats:
        - plain ``package_id``
        - simple package ID with version: ``package_id@version``
        - purl: ``pkg:npm/left-pad@3.7``

        If a specifier resolves to multiple constraints (as it might be the case for
        purl), we produce and returns all variations. That way the ``Solver`` below has
        all necessary details to resolve the constraints.

        Returns a generator of ``Specifier``.
        """
        specs = cls.parse_purl(spec_str)
        if specs:
            yield from specs

        # Specifier contains a version.
        elif VERSION_SEP in spec_str:
            package_id, version = spec_str.split(VERSION_SEP, 1)
            yield cls(
                raw_spec=spec_str,
                package_id=package_id,
                version=version,
            )

        # The spec is the plain package ID.
        else:
            yield cls(
                raw_spec=spec_str,
                package_id=spec_str,
            )

    @cached_property
    def parsed_version(self) -> TokenizedString:
        return parse_version(self.version)

    @cached_property
    def is_blank(self) -> bool:
        """Is considered blank a ``Specifier`` without any constraint on ``manager_id``
        or ``version``."""
        return not bool(self.manager_id or self.version)

    def __str__(self) -> str:
        """Human readable string of the spec.

        Dynamiccaly adds version, its separator and manage ID prefix (in purl syntax).
        """
        string = self.package_id
        if self.version:
            string = f"{string}{VERSION_SEP}{self.version}"
        if self.manager_id:
            string = f"pkg:{self.manager_id}/{string}"
        return string


class EmptyReduction(Exception):
    """Raised by the solver if no constraint can't be met."""


class Solver:
    """Combine a set of ``Specifier`` and allow for the solving of the constraints they
    represent."""

    manager_priority: Sequence[str] = []

    spec_pool: set[Specifier] = set()

    def __init__(
        self,
        spec_strings: Iterable[str] | None = None,
        manager_priority=None,
    ) -> None:
        if spec_strings:
            self.populate_from_strings(spec_strings)
        if manager_priority:
            self.manager_priority = manager_priority

    def populate_from_strings(self, spec_strings: Iterable[str]):
        """Populate the solver with package specifiers parsed from provided strings."""
        # Deduplicate entries.
        for spec_str in set(spec_strings):
            new_specs = Specifier.from_string(spec_str)
            self.spec_pool = self.spec_pool.union(new_specs)

    def top_priority_manager(
        self,
        keep_managers: Iterable[str | None] | None = None,
    ) -> str | None:
        """Returns the top priority manager configured on the solver.

        ``keep_managers`` allows for filtering by discarding managers not in that list.
        """
        for manager_id in self.manager_priority:
            # Returns the first manager in the priority list if no filtering needs to
            # be performed.
            if not keep_managers:
                return manager_id
            # Returns the first matching manager in the priority list.
            elif manager_id in keep_managers:
                return manager_id
        # No single top-priority manager can selected.
        return None

    def reduce_specs(self, specs: Iterable[Specifier]) -> Specifier:
        """Reduce a collection of ``Specifier`` to its essential, minimal and unique
        form.

        This method assumes that all provided ``specs`` are of the same package (like
        ``resolve_package_specs()`` does).

        The reduction process consist of several steps. At each step, as soon as we
        managed to reduce the constraints to one ``Specifier``, we returns it.

        Filtering steps:
        1. We remove all constraints tied to all by the top priority manager if
            provided.
        2. If no manager priority is provided, we discard constraints not tied to a
            manager.
        3. We discard constraints not tied to a version.
        4. We only keep constraints tied to the highest version.

        If we ends up with more than one set of constraints after all this filtering, an
        error is raised to invite the developer to troubleshoot the situation and refine
        this process.
        """
        # Deduplicate specifiers.
        collection = set(specs)

        if len(collection) == 1:
            return collection.pop()

        # If constraints allows for multiple managers, only keep specs matching
        # the highest priority.
        target_manager_ids = {s.manager_id for s in collection}
        if len(target_manager_ids) > 1:
            if self.manager_priority:
                if target_manager_ids.isdisjoint(self.manager_priority):
                    logging.warning(
                        f"Requested target managers {target_manager_ids} "
                        f"don't match selected {self.manager_priority}.",
                    )
                    raise EmptyReduction
                top_priority_manager = self.top_priority_manager(target_manager_ids)
                collection = {
                    s for s in collection if s.manager_id == top_priority_manager
                }

            # If no manager priority has been set, discards specs not tied to any
            # manager.
            else:
                collection = {s for s in collection if s.manager_id}

        if len(collection) == 1:
            return collection.pop()

        # Only keep the subset with the higher version.
        max_version = max(s.parsed_version for s in collection)
        collection = {s for s in collection if s.parsed_version == max_version}

        if len(collection) == 1:
            return collection.pop()

        # Still too much constraints.
        msg = (
            f"Cannot reduce {collection} any further. More heuristics must be "
            "implemented."
        )
        raise ValueError(msg)

    def resolve_package_specs(self) -> Iterator[tuple[str, Specifier]]:
        """Regroup specs of the pool by package IDs, and solve their constraints.

        Return each package ID with its single, reduced spec, or ``None`` if it ha no
        constraints.
        """
        # Regroup specs by package IDs. Has the nice side effect of deduplicating specs.
        keyfunc = attrgetter("package_id")
        for package_id, package_specs in groupby(
            sorted(self.spec_pool, key=keyfunc),
            key=keyfunc,
        ):
            # Serialize because of reuse in log message below.
            specs = tuple(package_specs)

            # Reduce and cleanup each set of constraints.
            try:
                reduced_spec = self.reduce_specs(specs)
            except EmptyReduction:
                logging.warning(f"Skip package {package_id}.")
                continue

            # Print warning if specifiers were subject to a reduction.
            if len(specs) > 1:
                logging.warning(
                    f"{package_id} specifiers reduced from "
                    f"{', '.join(sorted(s.raw_spec for s in specs))} to {reduced_spec}",
                )

            yield package_id, reduced_spec

    def resolve_specs_group_by_managers(self) -> dict[str | None, set[Specifier]]:
        """Resolves package specs, and returns them grouped by managers."""
        packages_per_managers: dict[str | None, set[Specifier]] = {}
        for _package_id, spec in self.resolve_package_specs():
            manager_id = None
            if spec:
                manager_id = spec.manager_id
            packages_per_managers.setdefault(manager_id, set()).add(spec)

        return packages_per_managers



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re
from typing import Iterator

from click_extra.platforms import ALL_PLATFORMS

from meta_package_manager.base import Package, PackageManager
from meta_package_manager.capabilities import (
    search_capabilities,
    version_not_implemented,
)
from meta_package_manager.version import parse_version


class Gem(PackageManager):
    """The RubyGems package manager.

    ..tip::

        Installs require ``sudo`` on system ruby. I (@tresni) recommend doing something
        like:

        .. code-block:: shell-session
            ► sudo dseditgroup -o edit -a -t user wheel

        And then do ``visudo`` to make it so the ``wheel`` group does not require
        a password. There is a line already there for it, you just need to
        uncomment it and save.
    """

    name = "Ruby Gems"

    homepage_url = "https://rubygems.org"

    platforms = ALL_PLATFORMS

    # Default to the version shipped with the latest maintained macOS version,
    # i.e. macOS 10.13 High Sierra, which is bundled with gem 2.5.2.
    requirement = "2.5.0"
    """
    .. code-block:: shell-session

        ► gem --version
        3.0.3
    """

    # Help mpm a little bit in its search for the `gem` binary.
    cli_search_path = ("/usr/local/opt/ruby/bin/gem", "/usr/local/opt/ruby/bin")

    post_args = ("--quiet",)  # Silence command progress meter

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            ► gem list --quiet
            bigdecimal (default: 1.4.1)
            bundler (default: 1.17.2)
            CFPropertyList (2.3.6)
            cmath (default: 1.0.0)
            csv (default: 3.0.9)
            date (default: 2.0.0)
            fileutils (1.4.1, default: 1.1.0)
            io-console (0.5.6, default: 0.4.7)
            ipaddr (default: 1.2.2)
            molinillo (0.5.4, 0.4.5, 0.2.3)
            nokogiri (1.5.6)
            psych (2.0.0)
            rake (0.9.6)
            rdoc (4.0.0)
            sqlite3 (1.3.7)
            test-unit (2.0.0.0)
        """
        output = self.run_cli("list")

        regexp = re.compile(r"(\S+) \((.+)\)")
        for package in output.splitlines():
            match = regexp.match(package)
            if match:
                package_id, versions = match.groups()
                # Guess latest installed version.
                version = max(
                    parse_version(v)
                    for v in re.compile(r",|default:| ").split(versions)
                    if v
                )
                yield self.package(id=package_id, installed_version=version)

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            ► gem outdated --quiet
            did_you_mean (1.0.0 < 1.0.2)
            io-console (0.4.5 < 0.4.6)
            json (1.8.3 < 2.0.1)
            minitest (5.8.3 < 5.9.0)
            power_assert (0.2.6 < 0.3.0)
            psych (2.0.17 < 2.1.0)
        """
        output = self.run_cli("outdated")

        regexp = re.compile(r"(\S+) \((\S+) < (\S+)\)")
        for package in output.splitlines():
            match = regexp.match(package)
            if match:
                package_id, current_version, latest_version = match.groups()
                yield self.package(
                    id=package_id,
                    installed_version=current_version,
                    latest_version=latest_version,
                )

    @search_capabilities(extended_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. caution::
            Search does not supports extended mode.

        .. code-block:: shell-session

            ► gem search python --versions --quiet
            at_coder_friends-generator-python_ref (0.2.0)
            bee_python (0.2.3)
            dependabot-python (0.117.5)
            logstash-filter-python (0.0.1 java)
            python (0.0.1)
            python-generator (1.1.0)
            python_with_git_test (2.499.8)
            rabbit-slide-niku-erlangvm-for-pythonista (2015.09.12)
            RubyToPython (0.0)

        .. code-block:: shell-session

            ► gem search python --versions --exact --quiet
            python (0.0.1)
        """
        search_arg = []
        if exact:
            search_arg.append("--exact")

        output = self.run_cli("search", query, "--versions", search_arg)

        regexp = re.compile(
            r"""
            (?P<package_id>\S+)     # Any string.
            \                       # A space.
            \(                      # Start of content in parenthesis.
                (?P<version>\S+)    # Version string.
                (?:\ \S+)?          # Optional platform value after space.
            \)
            """,
            re.VERBOSE,
        )

        for package_id, version in regexp.findall(output):
            yield self.package(id=package_id, latest_version=version)

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            ► gem install --user-install --quiet markdown
            Fetching kramdown-2.3.1.gem
            Fetching concurrent-ruby-1.1.9.gem
            (...)
            Fetching rubyzip-2.3.2.gem
            Fetching logutils-0.6.1.gem
            Fetching markdown-1.2.0.gem
            Successfully installed kramdown-2.3.1
            Successfully installed rubyzip-2.3.2
            (...)
            Successfully installed markdown-1.2.0
            (...)
            Parsing documentation for markdown-1.2.0
            Installing ri documentation for markdown-1.2.0
            Done installing documentation for (...) markdown after 19 seconds
            12 gems installed
        """
        return self.run_cli(
            "install",
            "--user-install",
            *self.post_args,
            package_id,
            auto_post_args=False,
        )

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            ► gem update --user-install --quiet
        """
        return self.build_cli(
            "update",
            "--user-install",
            *self.post_args,
            auto_post_args=False,
        )

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            ► gem update --user-install --quiet markdown
        """
        return self.build_cli(
            "update",
            "--user-install",
            *self.post_args,
            package_id,
            auto_post_args=False,
        )

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: shell-session

            ► gem uninstall --user-install --quiet left-pad
            Successfully uninstalled left-pad-1.1.0
        """
        return self.run_cli("uninstall", package_id)

    def cleanup(self) -> None:
        """Removes things we don't need anymore.

        .. code-block:: shell-session

            ► gem cleanup --quiet
            Cleaning up installed gems...
            Attempting to uninstall test-unit-3.2.9
            Unable to uninstall test-unit-3.2.9:
                Gem::FilePermissionError: You don't have write permissions \
                for the /Library/Ruby/Gems/2.6.0 directory.
            Attempting to uninstall did_you_mean-1.3.0
            Unable to uninstall did_you_mean-1.3.0:
                Gem::FilePermissionError: You don't have write permissions \
                for the /Library/Ruby/Gems/2.6.0 directory.
            Clean up complete
        """
        self.run_cli("cleanup")



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re

import pytest

from meta_package_manager.base import Operations

from .conftest import default_manager_ids
from .test_cli import CLISubCommandTests


@pytest.fixture()
def subcmd():
    return "upgrade", "--all"


class TestUpgrade(CLISubCommandTests):
    """Test the system-wide upgrade sub-command.

    .. danger::
        All tests here should me marked as destructive unless --dry-run parameter is
        passed.
    """

    @staticmethod
    def evaluate_signals(mid, stdout, stderr):
        yield from (
            f"warning: {mid} does not implement upgrade_all_cli." in stderr,
            f"warning: {mid} does not implement {Operations.upgrade_all}." in stderr,
            f"Upgrade all outdated packages from {mid}..." in stderr,
            bool(re.search(rf"Upgrade \S+ with {mid}\.\.\.", stderr)),
            # Common "not found" warning message.
            f"warning: Skip unavailable {mid} manager." in stderr,
        )

    @pytest.mark.parametrize("all_option", ("--all", None))
    def test_all_managers_dry_run_upgrade_all(self, invoke, all_option):
        result = invoke("--dry-run", "upgrade", all_option)
        assert result.exit_code == 0
        if not all_option:
            assert "assume -A/--all option" in result.stderr
        self.check_manager_selection(result)

    @pytest.mark.destructive()
    @pytest.mark.parametrize("all_option", ("--all", None))
    def test_all_managers_upgrade_all(self, invoke, all_option):
        result = invoke("upgrade", all_option)
        assert result.exit_code == 0
        if not all_option:
            assert "assume -A/--all option" in result.stderr
        self.check_manager_selection(result)

    @default_manager_ids
    @pytest.mark.parametrize("all_option", ("--all", None))
    def test_single_manager_dry_run_upgrade_all(self, invoke, manager_id, all_option):
        result = invoke(
            f"--{manager_id}", "--dry-run", "--verbosity", "INFO", "upgrade", all_option
        )
        assert result.exit_code == 0
        if not all_option:
            assert "assume -A/--all option" in result.stderr
        self.check_manager_selection(result, {manager_id})

    @pytest.mark.destructive()
    @default_manager_ids
    @pytest.mark.parametrize("all_option", ("--all", None))
    def test_single_manager_upgrade_all(self, invoke, manager_id, all_option):
        result = invoke(f"--{manager_id}", "upgrade", all_option)
        assert result.exit_code == 0
        if not all_option:
            assert "assume -A/--all option" in result.stderr
        self.check_manager_selection(result, {manager_id})



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Helpers and utilities to render and print content.

.. todo::

    Some of these are good candidates for upstream contribution to ``click.extra``.
"""

from __future__ import annotations

import builtins
import contextlib
import json
import logging
import sys
from functools import cached_property, partial
from io import StringIO
from operator import itemgetter
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Sequence
from unittest.mock import patch

from boltons.iterutils import flatten
from boltons.strutils import strip_ansi
from click_extra import echo, get_current_context, style
from click_extra.colorize import default_theme as theme
from click_extra.tabulate import output_formats
from tabulate import DataRow, TableFormat, tabulate

from .bar_plugin import MPMPlugin
from .pool import pool
from .version import TokenizedString

if TYPE_CHECKING:
    from collections import Counter

SORTABLE_FIELDS = {
    "manager_id",
    "manager_name",
    "package_id",
    "package_name",
    "version",
}
"""List of fields IDs allowed to be sorted."""


def colored_diff(a, b, style_common=None, style_a=None, style_b=None):
    """Highlight the most common left part between ``a`` and ``b`` strings and their
    trailing differences.

    Always returns 2 strings.
    """
    # Set defaults stling methods.
    style_common = partial(style, fg="bright_black")
    style_a = partial(style, fg="red")
    style_b = partial(style, fg="green")

    if isinstance(a, TokenizedString):
        a = str(a)
    if isinstance(b, TokenizedString):
        b = str(b)

    common_size = 0
    if a and b:
        while (min(len(a), len(b)) - 1) >= common_size and a[common_size] == b[
            common_size
        ]:
            common_size += 1

    # Styling of common and different parts.
    colored_a = ""
    colored_b = ""
    if common_size:
        colored_a = colored_b = style_common(a[:common_size])
    if a:
        colored_a += style_a(a[common_size:])
    if b:
        colored_b += style_b(b[common_size:])

    return colored_a, colored_b


output_formats = sorted([*output_formats, "json"])


def print_json(data):
    """Pretty-print Python data to JSON and output results to ``<stdout>``.

    Serialize :py:class:`pathlib.Path` and
    :py:class:`meta_package_manager.version.TokenizedString` objects.
    """

    def serialize_objects(obj):
        if isinstance(obj, (TokenizedString, Path)):
            return str(obj)
        raise TypeError(repr(obj) + " is not JSON serializable.")

    echo(
        json.dumps(
            data,
            sort_keys=True,
            indent=4,
            separators=(",", ": "),
            default=serialize_objects,
        ),
        # Do not pollute output with ANSI codes.
        color=False,
    )


def print_table(
    header_defs: list[tuple[str, str]],
    rows: Iterable[Sequence[str | TokenizedString]],
    sort_key: str | None = None,
) -> None:
    """Print a table.

    ``header_defs`` parameter is an ordered list of tuple whose first item is the
    column's label and the second the column's ID. Example:

    .. code-block:: python

        [
            ("Column 1", "column1"),
            ("User's name", "name"),
            ("Package manager", "manager_id"),
            ...,
        ]

    Rows can be sorted by providing the column's ID to ``sort_key`` parameter. By
    default, ``None`` means the table will be sorted in the order of columns provided by
    ``header_defs``.
    """
    # Do not print anything, not even table headers if no rows.
    if not rows:
        return

    header_labels = (style(label, bold=True) for label, _ in header_defs)

    # Check there is no duplicate column IDs.
    header_ids = [col_id for _, col_id in header_defs if col_id]
    assert len(header_ids) == len(set(header_ids))

    # Default sorting follows the order of headers.
    sort_order = list(range(len(header_defs)))

    # Move the sorting key's index in the front of priority.
    if sort_key and sort_key in header_ids:
        # Build an index of column id's position.
        col_index = {col_id: i for i, (_, col_id) in enumerate(header_defs) if col_id}
        sort_column_index = col_index[sort_key]
        sort_order.remove(sort_column_index)
        sort_order.insert(0, sort_column_index)

    def sort_method(line):
        """Serialize line's content for natural sorting.

        1. Extract each cell value in the order provided by `sort_order`;
        2. Strip terminal color formatting;
        3. Then tokenize each cell's content for user-friendly natural sorting.
        """
        sorting_key = []
        for cell in itemgetter(*sort_order)(line):
            if isinstance(cell, TokenizedString):
                key = cell
            elif not cell:
                key = None
            else:
                key = TokenizedString(strip_ansi(cell))
            sorting_key.append(key)
        return tuple(sorting_key)

    ctx = get_current_context()
    ctx.find_root().print_table(  # type: ignore[attr-defined]
        sorted(rows, key=sort_method),
        header_labels,
    )


def print_stats(manager_stats: Counter) -> None:
    """Prints statistics to ``<stderr>``: total packages and a break down by package
    manager.

    Prints something like:

    .. code-block:: text

        10 packages total (brew: 2, pip: 2, gem: 2, vscode: 2, npm: 2, composer: 0).
    """
    per_manager_totals = ""
    if manager_stats:
        per_manager_totals = (
            f" ({', '.join(f'{k}: {v}' for k, v in manager_stats.most_common())})"
        )
    if sys.version_info >= (3, 10):
        total = manager_stats.total()
    else:
        total = sum(manager_stats.values())
    plural = "s" if total > 1 else ""
    echo(f"{total} package{plural} total{per_manager_totals}.", err=True)


class BarPluginRenderer(MPMPlugin):
    """All utilities used to render output compatible with both Xbar and SwiftBar plugin
    dialect.

    The minimal code to locate ``mpm``, then call it and print its output resides in the
    plugin itself at :py:meth:`meta_package_manager.bar_plugin.MPMPlugin.mpm_exec`.

    All other stuff, especially the rendering code, is managed here, to allow for more
    complex layouts relying on external Python dependencies. This also limits the number
    of required updates on the plugin itself.
    """

    @cached_property
    def submenu_layout(self) -> bool:
        """Group packages into manager sub-menus.

        If ``True``, will replace the default flat layout with an alternative structure
        where actions are grouped into submenus, one for each manager.

        Value is sourced from the ``VAR_SUBMENU_LAYOUT`` environment variable.
        """
        return self.getenv_bool("VAR_SUBMENU_LAYOUT", False)

    @cached_property
    def dark_mode(self) -> bool:
        """Detect dark mode by inspecting environment variables.

        Value is sourced from two environment variables depending on the plugin:

        - ``OS_APPEARANCE`` for SwiftBar
        - ``XBARDarkMode`` for XBar
        """
        if self.is_swiftbar:
            return self.getenv_str("OS_APPEARANCE", "light") == "dark"
        return self.getenv_bool("XBARDarkMode")

    @staticmethod
    def render_cli(cmd_args: tuple[str | Path, ...]) -> str:
        """Return a formatted CLI compatible with Xbar and SwiftBar plugin format.

        I.e. a string with this schema:

        .. code-block::

            shell=cmd_args[0] param1=cmd_args[1] param2=cmd_args[2] ...
        """
        plugin_params = []
        # Serialize Path into string.
        for index, param_value in enumerate(map(str, flatten(cmd_args))):
            param_id = "shell" if index == 0 else f"param{index}"
            plugin_params.append(f"{param_id}={param_value}")
        return " ".join(plugin_params)

    def print_cli_item(self, *args) -> None:
        """Print two CLI entries:

        - one that is silent
        - a second one that is the exact copy of the above but forces the execution
          by the way of a visible terminal
        """
        self.pp(*args, "terminal=false")
        self.pp(*args, "terminal=true", "alternate=true")

    def print_upgrade_all_item(self, manager: dict, submenu: str = "") -> None:
        """Print the menu entry to upgrade all outdated package of a manager."""
        if manager.get("upgrade_all_cli"):
            if self.submenu_layout:
                print("-----")
            self.print_cli_item(
                f"{submenu}🆙 Upgrade all {manager['id']} packages",
                manager["upgrade_all_cli"],
                self.default_font,
                "refresh=true",
            )

    plain_table_format = TableFormat(
        lineabove=None,
        linebelowheader=None,
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("", " ", ""),
        datarow=DataRow("", " ", ""),
        padding=0,
        with_header_hide=None,
    )
    """Simple rendering format with single-space separated columns used in the function
    below."""

    @staticmethod
    def render_table(
        table_data: Sequence[Sequence[str]] | None,
    ) -> Any | list:
        """Renders a table data with pre-configured alignment centered around the third
        column.

        Returns a list of strings, one item per line.

        .. code-block:: pycon

            >>> table_data = [
            ...     ("xmlrpc", "0.3.1", "→", "0.4"),
            ...     ("blockblock", "5.33,VHSDGataYCcV8xqv5TSZA", "→", "5.39"),
            ...     ("sed", "2", "→", "2021.0328"),
            ... ]
            >>> print(render_table(table_data))
            xmlrpc                          0.3.1 → 0.4
            blockblock 5.33,VHSDGataYCcV8xqv5TSZA → 5.39
            sed                                 2 → 2021.0328

        ..todo::

            Use upcoming ``tabulate.SEPARATING_LINE`` to produce the whole bar plugin
            table in one go and have all version numbers from all managers aligned. See:
            https://github.com/astanin/python-tabulate/commit/dab256d1f64da97720c1459478a3cc0a4ea7a91e
        """
        if not table_data:
            return []
        return tabulate(
            table_data,
            tablefmt=BarPluginRenderer.plain_table_format,
            colalign=("left", "right", "center", "left"),
            disable_numparse=True,
        ).splitlines()

    def _render(self, outdated_data) -> None:
        """Main method implementing the final structured rendering in *Bar plugin
        dialect.

        ..todo::

            Wait for ANSI-aware layout in table to be merged upstream so we can highly
            version differences in bar plugin. See:
            https://github.com/astanin/python-tabulate/pull/184
        """
        managers = outdated_data.values()
        font = self.monospace_font if self.table_rendering else self.default_font

        # Print menu bar icon with number of available upgrades.
        total_outdated = sum(len(m["packages"]) for m in managers)
        total_errors = sum(len(m.get("errors", [])) for m in managers)
        self.pp(
            (f"🎁↑{total_outdated}" if total_outdated else "📦✓")
            + (f" ⚠️{total_errors}" if total_errors else ""),
            "dropdown=false",
        )

        # Prefix for section content.
        submenu = "--" if self.submenu_layout else ""

        for manager in managers:
            package_count = len(manager["packages"])
            plural = "s" if package_count > 1 else ""
            package_label = f"package{plural}"

            table = [
                (
                    (
                        p.get("name") or p.get("id"),
                        p["installed_version"],
                        "→",
                        p["latest_version"],
                    ),
                    p["upgrade_cli"],
                )
                for p in manager["packages"]
            ]

            # Table-like rendering
            if self.table_rendering:
                header = f"{manager['id']} - {package_count} {package_label}"
                formatted_lines = self.render_table([p[0] for p in table])

            # Variable-width / non-table / non-monospaced rendering.
            else:
                header = f"{package_count} outdated {manager['name']} {package_label}"
                formatted_lines = [" ".join(map(str, p[0])) for p in table]

            upgrade_cli_list = [p[1] for p in table]

            assert len(formatted_lines) == len(upgrade_cli_list)

            # Print section separator before printing the manager header.
            print("---")

            # Print section header.
            error = ""
            if self.submenu_layout and manager.get("errors", None):
                error = "⚠️ "
            self.pp(f"{error}{header}", font)

            # Print a menu entry for each outdated packages.
            for line, upgrade_cli in zip(formatted_lines, upgrade_cli_list):
                self.print_cli_item(
                    f"{submenu}{line}",
                    upgrade_cli,
                    font,
                    "refresh=true",
                )

            self.print_upgrade_all_item(manager, submenu)

            for error_msg in manager.get("errors", []):
                print("-----" if self.submenu_layout else "---")
                self.print_error(error_msg, submenu)

    def render(self, outdated_data) -> str:
        """Wraps the :py:meth:`meta_package_manager.output.BarPluginRenderer._render`
        function above to capture all ``print`` statements."""
        capture = StringIO()
        print_capture = partial(print, file=capture)
        with patch.object(builtins, "print", new=print_capture):
            self._render(outdated_data)
        return capture.getvalue()

    def add_upgrade_cli(self, outdated_data):
        """Augment the outdated data from ``mpm outdated`` subcommand with upgrade CLI
        fields for bar plugin consumption."""
        for manager_id, manager_data in outdated_data.items():
            if manager_data.get("packages"):
                manager = pool.get(manager_id)

                # Produce the full-upgrade CLI.
                try:
                    upgrade_all_cli = manager.upgrade_all_cli()
                except NotImplementedError:
                    # Fallback on mpm itself which is capable of simulating a full
                    # upgrade.
                    logging.warning(
                        f"{theme.invoked_command(manager_id)} "
                        "does not implement upgrade_all_cli.",
                    )
                    upgrade_all_cli = (
                        *self.mpm_exec,
                        f"--{manager_id}",
                        "upgrade",
                        "--all",
                    )
                    logging.debug(f"Fallback to direct mpm call: {upgrade_all_cli}")

                # Update outdated data with the full-upgrade CLI.
                outdated_data[manager_id]["upgrade_all_cli"] = self.render_cli(
                    upgrade_all_cli,
                )

                # Add for each package its upgrade CLI.
                for package in manager_data["packages"]:
                    # Generate the version-less upgrade CLI to be used by the *bar
                    # plugin.
                    upgrade_cli = None
                    with contextlib.suppress(NotImplementedError):
                        upgrade_cli = self.render_cli(
                            manager.upgrade_one_cli(package["id"]),
                        )

                    package["upgrade_cli"] = upgrade_cli

        return outdated_data

    def print(self, outdated_data) -> None:
        """Print the final plugin rendering to ``<stdout>``.

        Capturing the output of the plugin and re-printing it will introduce an extra
        line return, hence the extra call to ``rstrip()``.
        """
        outdated_data = self.add_upgrade_cli(outdated_data)
        echo(self.render(outdated_data).rstrip())
